import errno
import hashlib
import os
import sys

import stat

from humanize import naturalsize
from qrcode import QRCode
from tqdm import tqdm
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.protocols import basic
from twisted.python import log
from wormhole import __version__, create

from ..errors import TransferError, UnsendableFileError
from .._status import WormholeStatus, ConsumedCode
from ..transit import TransitSender
from ..util import bytes_to_dict, bytes_to_hexstr, dict_to_bytes
from .welcome import handle_welcome

from iterableio import open_iterable
from zipstream.ng import ZipStream, walk

APPID = "lothar.com/wormhole/text-or-file-xfer"
VERIFY_TIMER = float(os.environ.get("_MAGIC_WORMHOLE_TEST_VERIFY_TIMER", 1.0))


def send(args, reactor=reactor):
    """I implement 'wormhole send'. I return a Deferred that fires with None
    (for success), or signals one of the following errors:
    * WrongPasswordError: the two sides didn't use matching passwords
    * Timeout: something didn't happen fast enough for our tastes
    * TransferError: the receiver rejected the transfer: verifier mismatch,
                     permission not granted, ack not successful.
    * any other error: something unexpected happened
    """
    return Sender(args, reactor).go()


class Sender:
    def __init__(self, args, reactor):
        self._args = args
        self._reactor = reactor
        self._tor = None
        self._timing = args.timing
        self._fd_to_send = None
        self._transit_sender = None
        self._status = WormholeStatus()

    @inlineCallbacks
    def go(self):
        assert isinstance(self._args.relay_url, str)
        if self._args.tor:
            with self._timing.add("import", which="tor_manager"):
                from ..tor_manager import get_tor
            # For now, block everything until Tor has started. Soon: launch
            # tor in parallel with everything else, make sure the Tor object
            # can lazy-provide an endpoint, and overlap the startup process
            # with the user handing off the wormhole code
            self._tor = yield get_tor(
                reactor,
                self._args.launch_tor,
                self._args.tor_control_port,
                timing=self._timing)

        w = create(
            self._args.appid or APPID,
            self._args.relay_url,
            self._reactor,
            tor=self._tor,
            timing=self._timing,
            on_status_update=self._on_status,
        )
        if self._args.debug_state:
            w.debug_set_trace("send", which=" ".join(self._args.debug_state), file=self._args.stdout)
        d = self._go(w)

        # if we succeed, we should close and return the w.close results
        # (which might be an error)
        @inlineCallbacks
        def _good(res):
            yield w.close()  # wait for ack
            return res

        # if we raise an error, we should close and then return the original
        # error (the close might give us an error, but it isn't as important
        # as the original one)
        @inlineCallbacks
        def _bad(f):
            try:
                yield w.close()  # might be an error too
            except Exception:
                pass
            return f

        d.addCallbacks(_good, _bad)
        yield d

    def _send_data(self, data, w):
        data_bytes = dict_to_bytes(data)
        w.send_message(data_bytes)

    def _on_status(self, status):
        if not isinstance(self._status.code, ConsumedCode) and isinstance(status.code, ConsumedCode):
            print("Note: code has been consumed and can no longer be used.", file=self._args.stdout)
        self._status = status

    @inlineCallbacks
    def _go(self, w):
        welcome = yield w.get_welcome()
        handle_welcome(welcome, self._args.relay_url, __version__,
                       self._args.stderr)

        # TODO: run the blocking zip-the-directory IO in a thread, let the
        # wormhole exchange happen in parallel
        offer, self._fd_to_send = self._build_offer()
        args = self._args

        other_cmd = "wormhole receive"
        if args.verify:
            other_cmd = "wormhole receive --verify"
        if args.zeromode:
            assert not args.code
            args.code = "0-"
            other_cmd += " -0"

        if args.code:
            w.set_code(args.code)
        else:
            w.allocate_code(args.code_length)

        code = yield w.get_code()
        if not args.zeromode:
            print(f"Wormhole code is: {code}", file=args.stderr)
            other_cmd += " " + code

        if not args.zeromode and args.qr:
            qr = QRCode(border=1)
            qr.add_data(f"wormhole-transfer:{code}")
            # Use TTY colors if available, otherwise use ASCII
            if args.stderr.isatty():
                qr.print_ascii(out=args.stderr, tty=True, invert=False)
            else:
                qr.print_ascii(out=args.stderr, tty=False, invert=False)

        print("On the other computer, please run:", file=args.stderr)
        print("", file=args.stderr)
        print(other_cmd, file=args.stderr)
        print("", file=args.stderr)
        # flush stderr so the code is displayed immediately
        args.stderr.flush()

        # We don't print a "waiting" message for get_unverified_key() here,
        # even though we do that in cmd_receive.py, because it's not at all
        # surprising to we waiting here for a long time. We'll sit in
        # get_unverified_key() until the receiver has typed in the code and
        # their PAKE message makes it to us.
        yield w.get_unverified_key()

        # TODO: don't stall on w.get_verifier() unless they want it
        def on_slow_connection():
            print(
                "Key established, waiting for confirmation...",
                file=args.stderr)

        notify = self._reactor.callLater(VERIFY_TIMER, on_slow_connection)
        try:
            # The usual sender-chooses-code sequence means the receiver's
            # PAKE should be followed immediately by their VERSION, so
            # w.get_verifier() should fire right away. However if we're
            # using the offline-codes sequence, and the receiver typed in
            # their code first, and then they went offline, we might be
            # sitting here for a while, so printing the "waiting" message
            # seems like a good idea. It might even be appropriate to give up
            # after a while.
            verifier_bytes = yield w.get_verifier()  # might WrongPasswordError
        finally:
            if not notify.called:
                notify.cancel()

        if args.verify:
            # check_verifier() does a blocking call to input(), so stall for
            # a moment to let any outbound messages get written into the
            # kernel. At this point, we're sitting in a callback of
            # get_verifier(), which is triggered by receipt of the other
            # side's VERSION message. But we might have gotten both the PAKE
            # and the VERSION message in the same turn, and our outbound
            # VERSION message (triggered by receipt of their PAKE) is still
            # in Twisted's transmit queue. If we don't wait a moment, it will
            # be stuck there until `input()` returns, and the receiver won't
            # be able to compute a Verifier for the users to compare. #349
            # has more details
            d = Deferred()
            reactor.callLater(0.001, d.callback, None)
            yield d
            self._check_verifier(w,
                                 verifier_bytes)  # blocks, can TransferError

        if self._fd_to_send:
            ts = TransitSender(
                args.transit_helper,
                no_listen=(not args.listen),
                tor=self._tor,
                reactor=self._reactor,
                timing=self._timing)
            self._transit_sender = ts

            # for now, send this before the main offer
            sender_abilities = ts.get_connection_abilities()
            sender_hints = yield ts.get_connection_hints()
            sender_transit = {
                "abilities-v1": sender_abilities,
                "hints-v1": sender_hints,
            }
            self._send_data({"transit": sender_transit}, w)

            # When I made it possible to override APPID with a CLI argument
            # (issue #113), I forgot to also change this w.derive_key()
            # (issue #339). We're stuck with it now. Use a local constant to
            # make this clear.
            BUG339_APPID = "lothar.com/wormhole/text-or-file-xfer"

            # TODO: move this down below w.get_message()
            transit_key = w.derive_key(BUG339_APPID + "/transit-key",
                                       ts.TRANSIT_KEY_LENGTH)
            ts.set_transit_key(transit_key)

        self._send_data({"offer": offer}, w)

        want_answer = True

        while True:
            them_d_bytes = yield w.get_message()
            # TODO: get_message() fired, so get_verifier must have fired, so
            # now it's safe to use w.derive_key()
            them_d = bytes_to_dict(them_d_bytes)
            # print("GOT", them_d)
            recognized = False
            if "error" in them_d:
                raise TransferError(
                    f"remote error, transfer abandoned: {them_d['error']}")
            if "transit" in them_d:
                recognized = True
                yield self._handle_transit(them_d["transit"])
            if "answer" in them_d:
                recognized = True
                if not want_answer:
                    raise TransferError("duplicate answer")
                want_answer = True
                yield self._handle_answer(them_d["answer"])
                return None
            if not recognized:
                log.msg(f"unrecognized message {them_d!r}")

    def _check_verifier(self, w, verifier_bytes):
        verifier = bytes_to_hexstr(verifier_bytes)
        while True:
            ok = input(f"Verifier {verifier}. ok? (yes/no): ")
            if ok.lower() == "yes":
                break
            if ok.lower() == "no":
                err = "sender rejected verification check, abandoned transfer"
                reject_data = dict_to_bytes({"error": err})
                w.send_message(reject_data)
                raise TransferError(err)

    def _handle_transit(self, receiver_transit):
        ts = self._transit_sender
        ts.add_connection_hints(receiver_transit.get("hints-v1", []))

    def _build_offer(self):
        offer = {}

        args = self._args
        text = args.text
        if text == "-":
            print("Reading text message from stdin..", file=args.stderr)
            text = sys.stdin.read()
        if not text and not args.what:
            text = input("Text to send: ")

        if text is not None:
            print(
                f"Sending text message ({naturalsize(len(text))})",
                file=args.stderr)
            offer = {"message": text}
            fd_to_send = None
            return offer, fd_to_send

        # click.Path (with resolve_path=False, the default) does not do path
        # resolution, so we must join it to cwd ourselves. We could use
        # resolve_path=True, but then it would also do os.path.realpath(),
        # which would replace the basename with the target of a symlink (if
        # any), which is not what I think users would want: if you symlink
        # X->Y and send X, you expect the recipient to save it in X, not Y.
        #
        # TODO: an open question is whether args.cwd (i.e. os.getcwd()) will
        # be unicode or bytes. We need it to be something that can be
        # os.path.joined with the unicode args.what .
        what = os.path.join(args.cwd, args.what)

        # We always tell the receiver to create a file (or directory) with the
        # same basename as what the local user typed, even if the local object
        # is a symlink to something with a different name. The normpath() is
        # there to remove trailing slashes.
        basename = os.path.basename(os.path.normpath(what))
        assert basename != "", what  # normpath shouldn't allow this

        # We use realpath() instead of normpath() to locate the actual
        # file/directory, because the path might contain symlinks, and
        # normpath() would collapse those before resolving them.
        # test_cli.OfferData.test_symlink_collapse tests this.

        # Unfortunately on windows, realpath() (on py3) is built out of
        # normpath() because of a py2-era belief that windows lacks a working
        # os.path.islink(): see https://bugs.python.org/issue9949 . The
        # consequence is that "wormhole send PATH" might send the wrong file,
        # if:
        # * we're running on windows
        # * PATH goes down through a symlink and then up with parent-directory
        #   navigation (".."), then back down again
        # * the back-down-again portion of the path also exists under the
        #   original directory (an error is thrown if not)

        # I'd like to fix this. The core issue is sending directories with a
        # trailing slash: we need to 1: open the right directory, and 2: strip
        # the right parent path out of the filenames we get from os.walk(). We
        # used to use what.rstrip() for this, but bug #251 reported this
        # failing on windows-with-bash. realpath() works in both those cases,
        # but fails with the up-down symlinks situation. I think we'll need to
        # find a third way to strip the trailing slash reliably in all
        # environments.

        what = os.path.realpath(what)
        if not os.path.exists(what):
            raise TransferError(
                f"Cannot send: no file/directory named '{args.what}'")

        if os.path.isfile(what):
            # we're sending a file
            filesize = os.stat(what).st_size
            offer["file"] = {
                "filename": basename,
                "filesize": filesize,
            }
            print(
                f"Sending {naturalsize(filesize)} file named '{basename}'",
                file=args.stderr)
            fd_to_send = open(what, "rb")
            return offer, fd_to_send

        if os.path.isdir(what):
            print("Building zipfile..", file=args.stderr)
            # We're sending a directory, stream it as a zipfile

            zs = ZipStream(sized=True)
            for filepath in walk(what, preserve_empty=True, followlinks=True):
                try:
                    if not os.access(filepath, os.R_OK):
                        raise PermissionError(errno.EACCES, os.strerror(errno.EACCES), filepath)
                    zs.add_path(
                        filepath,
                        arcname=os.path.relpath(filepath, what),
                        recurse=False,
                    )
                except OSError as e:
                    errmsg = f"{filepath}: {e.strerror}"
                    if not self._args.ignore_unsendable_files:
                        raise UnsendableFileError(errmsg)
                    print(
                        f"{errmsg} (ignoring error)",
                        file=args.stderr
                    )

            filesizes = [x["size"] for x in zs.info_list() if not x["is_dir"]]
            filesize = len(zs)
            offer["directory"] = {
                "mode": "zipfile/deflated",
                "dirname": basename,
                "zipsize": filesize,
                "numbytes": sum(filesizes),
                "numfiles": len(filesizes),
            }
            print(
                "Sending directory (%s compressed) named '%s'" %
                (naturalsize(filesize), basename),
                file=args.stderr)
            return offer, zs

        if stat.S_ISBLK(os.stat(what).st_mode):
            fd_to_send = open(what, "rb")
            filesize = fd_to_send.seek(0, 2)

            offer["file"] = {
                "filename": basename,
                "filesize": filesize,
            }
            print(
                f"Sending {naturalsize(filesize)} block device named '{basename}'",
                file=args.stderr)

            fd_to_send.seek(0)
            return offer, fd_to_send

        raise TypeError(f"'{args.what}' is neither file nor directory")

    @inlineCallbacks
    def _handle_answer(self, them_answer):
        if self._fd_to_send is None:
            if them_answer["message_ack"] == "ok":
                print("text message sent", file=self._args.stderr)
                return None
            raise TransferError(f"error sending text: {them_answer!r}")

        if them_answer.get("file_ack") != "ok":
            raise TransferError("ambiguous response from remote, "
                                "transfer abandoned: %s" % (them_answer, ))

        yield self._send_file()

    @inlineCallbacks
    def _send_file(self):
        ts = self._transit_sender

        if isinstance(self._fd_to_send, ZipStream):
            filesize = len(self._fd_to_send)
            self._fd_to_send = open_iterable(self._fd_to_send, "rb")
        else:
            self._fd_to_send.seek(0, 2)
            filesize = self._fd_to_send.tell()
            self._fd_to_send.seek(0, 0)

        record_pipe = yield ts.connect()
        self._timing.add("transit connected")
        # record_pipe should implement IConsumer, chunks are just records
        stderr = self._args.stderr
        print(f"Sending ({record_pipe.describe()})..", file=stderr)

        hasher = hashlib.sha256()
        progress = tqdm(
            file=stderr,
            disable=self._args.hide_progress,
            unit="B",
            unit_scale=True,
            dynamic_ncols=True,
            total=filesize)

        def _count_and_hash(data):
            hasher.update(data)
            progress.update(len(data))
            return data

        fs = basic.FileSender()

        with self._timing.add("tx file"):
            with progress:
                if filesize:
                    # don't send zero-length files
                    yield fs.beginFileTransfer(
                        self._fd_to_send,
                        record_pipe,
                        transform=_count_and_hash)

        expected_hash = hasher.digest()
        expected_hex = bytes_to_hexstr(expected_hash)
        print("File sent.. waiting for confirmation", file=stderr)
        with self._timing.add("get ack") as t:
            ack_bytes = yield record_pipe.receive_record()
            record_pipe.close()
            ack = bytes_to_dict(ack_bytes)
            ok = ack.get("ack", "")
            if ok != "ok":
                t.detail(ack="failed")
                raise TransferError(f"Transfer failed (remote says: {ack!r})")
            if "sha256" in ack:
                if ack["sha256"] != expected_hex:
                    t.detail(datahash="failed")
                    raise TransferError("Transfer failed (bad remote hash)")
            print("Confirmation received. Transfer complete.", file=stderr)
            t.detail(ack="ok")



import struct

assert len(struct.pack("<L", 0)) == 4
assert len(struct.pack("<Q", 0)) == 8


def to_be4(value):
    if not 0 <= value < 2**32:
        raise ValueError
    return struct.pack(">L", value)


def from_be4(b):
    if not isinstance(b, bytes):
        raise TypeError(repr(b))
    if len(b) != 4:
        raise ValueError
    return struct.unpack(">L", b)[0]



import os
import time
start = time.time()

from sys import stderr, stdout  # noqa: E402
from textwrap import dedent, fill  # noqa: E402

import click  # noqa: E402
from twisted.internet.defer import inlineCallbacks, maybeDeferred  # noqa: E402
from twisted.internet.task import react  # noqa: E402
from twisted.python.failure import Failure  # noqa: E402

from . import public_relay  # noqa: E402
from .. import __version__  # noqa: E402
from ..errors import (KeyFormatError, NoTorError,  # noqa: E402
                      ServerConnectionError,
                      TransferError, UnsendableFileError, WelcomeError,
                      WrongPasswordError)
from ..timing import DebugTiming  # noqa: E402

top_import_finish = time.time()


class Config:
    """
    Union of config options that we pass down to (sub) commands.
    """

    def __init__(self):
        # This only holds attributes which are *not* set by CLI arguments.
        # Everything else comes from Click decorators, so we can be sure
        # we're exercising the defaults.
        self.timing = DebugTiming()
        self.cwd = os.getcwd()
        self.stdout = stdout
        self.stderr = stderr
        self.tor = False  # XXX?
        self._debug_state = None

    @property
    def debug_state(self):
        return self._debug_state

    @debug_state.setter
    def debug_state(self, debug_state):
        if not debug_state:
            return
        valid_machines = [
            'B', 'N', 'M', 'S', 'O', 'K', 'SK', 'R', 'RC', 'L', 'C', 'T'
        ]
        debug_state = debug_state.split(",")
        invalid_machines = [
            machine
            for machine in debug_state
            if machine not in valid_machines
        ]
        if invalid_machines:
            raise click.UsageError(
                f"Cannot debug unknown machines: {' '.join(invalid_machines)}"
            )
        self._debug_state = debug_state


def _compose(*decorators):
    def decorate(f):
        for d in reversed(decorators):
            f = d(f)
        return f

    return decorate


ALIASES = {
    "tx": "send",
    "rx": "receive",
    "recieve": "receive",
    "recv": "receive",
}


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        cmd_name = ALIASES.get(cmd_name, cmd_name)
        return click.Group.get_command(self, ctx, cmd_name)


# top-level command ("wormhole ...")
@click.group(cls=AliasedGroup)
@click.option("--appid", default=None, metavar="APPID", help="appid to use")
@click.option(
    "--relay-url",
    default=public_relay.RENDEZVOUS_RELAY,
    envvar='WORMHOLE_RELAY_URL',
    show_envvar=True,
    metavar="URL",
    help="rendezvous relay to use",
)
@click.option(
    "--transit-helper",
    default=public_relay.TRANSIT_RELAY,
    envvar='WORMHOLE_TRANSIT_HELPER',
    show_envvar=True,
    metavar="tcp:HOST:PORT",
    help="transit relay to use",
)
@click.option(
    "--dump-timing",
    type=str,  # TODO: hide from --help output
    default=None,
    metavar="FILE.json",
    help="(debug) write timing data to file",
)
@click.version_option(
    message="magic-wormhole %(version)s",
    version=__version__,
)
@click.pass_context
def wormhole(context, dump_timing, transit_helper, relay_url, appid):
    """
    Create a Magic Wormhole and communicate through it.

    Wormholes are created by speaking the same magic CODE in two
    different places at the same time.  Wormholes are secure against
    anyone who doesn't use the same code.
    """
    context.obj = cfg = Config()
    cfg.appid = appid
    cfg.relay_url = relay_url
    cfg.transit_helper = transit_helper
    cfg.dump_timing = dump_timing


@inlineCallbacks
def _dispatch_command(reactor, cfg, command):
    """
    Internal helper. This calls the given command (a no-argument
    callable) with the Config instance in cfg and interprets any
    errors for the user.
    """
    cfg.timing.add("command dispatch")
    cfg.timing.add(
        "import", when=start, which="top").finish(when=top_import_finish)

    try:
        yield maybeDeferred(command)
    except (WrongPasswordError, NoTorError) as e:
        msg = fill("ERROR: " + dedent(e.__doc__))
        print(msg, file=cfg.stderr)
        raise SystemExit(1)
    except (WelcomeError, UnsendableFileError, KeyFormatError) as e:
        msg = fill("ERROR: " + dedent(e.__doc__))
        print(msg, file=cfg.stderr)
        print("", file=cfg.stderr)
        print(str(e), file=cfg.stderr)
        raise SystemExit(1)
    except TransferError as e:
        print(f"TransferError: {str(e)}", file=cfg.stderr)
        raise SystemExit(1)
    except ServerConnectionError as e:
        msg = fill("ERROR: " + dedent(e.__doc__)) + "\n"
        msg += f"(relay URL was {e.url})\n"
        msg += str(e)
        print(msg, file=cfg.stderr)
        raise SystemExit(1)
    except Exception as e:
        # this prints a proper traceback, whereas
        # traceback.print_exc() just prints a TB to the "yield"
        # line above ...
        Failure().printTraceback(file=cfg.stderr)
        print("ERROR:", str(e), file=cfg.stderr)
        raise SystemExit(1)

    cfg.timing.add("exit")
    if cfg.dump_timing:
        cfg.timing.write(cfg.dump_timing, cfg.stderr)


CommonArgs = _compose(
    click.option(
        "-0",
        "zeromode",
        default=False,
        is_flag=True,
        help="enable no-code anything-goes mode",
    ),
    click.option(
        "-c",
        "--code-length",
        default=None,
        metavar="NUMWORDS",
        help="length of code (in bytes/words)",
    ),
    click.option(
        "-v",
        "--verify",
        is_flag=True,
        default=False,
        help="display verification string (and wait for approval)",
    ),
    click.option(
        "--hide-progress",
        is_flag=True,
        default=False,
        help="suppress progress-bar display",
    ),
    click.option(
        "--listen/--no-listen",
        default=True,
        help="(debug) don't open a listening socket for Transit",
    ),
)

TorArgs = _compose(
    click.option(
        "--tor",
        is_flag=True,
        default=False,
        help="use Tor when connecting",
    ),
    click.option(
        "--launch-tor",
        is_flag=True,
        default=False,
        help="launch Tor, rather than use existing control/socks port",
    ),
    click.option(
        "--tor-control-port",
        default=None,
        metavar="ENDPOINT",
        help="endpoint descriptor for Tor control port",
    ),
)


@wormhole.command()
@click.pass_context
def help(context, **kwargs):
    print(context.find_root().get_help())


# wormhole send (or "wormhole tx")
@wormhole.command()
@CommonArgs
@TorArgs
@click.option(
    "--code",
    metavar="CODE",
    help="human-generated code phrase",
)
@click.option(
    "--text",
    default=None,
    metavar="MESSAGE",
    help=("text message to send, instead of a file."
          " Use '-' to read from stdin."),
)
@click.option(
    "--ignore-unsendable-files",
    default=False,
    is_flag=True,
    help="Don't raise an error if a file can't be read.")
@click.option(
    "--debug-state",
    is_flag=False,
    flag_value="B,N,M,S,O,K,SK,R,RC,L,C,T",
    default=None,
    metavar="MACHINES",
    help=(
        "Debug state-machine transitions. "
        "Possible machines to debug are accepted as a comma-separated list "
        "and the default is all of them. Valid machines are "
        "any of: B,N,M,S,O,K,SK,R,RC,L,C,T"
    )
)
@click.option(
    "--qr/--no-qr",
    default=True,
    help="Generate and show ASCII-based QR code.",
    envvar="WORMHOLE_QR",
    show_envvar=True,
)
@click.argument("what", required=False, type=click.Path(path_type=str))
@click.pass_obj
def send(cfg, **kwargs):
    """Send a text message, file, or directory"""
    kwargs["code_length"] = 2 if kwargs["code_length"] is None else int(kwargs["code_length"])

    for name, value in kwargs.items():
        setattr(cfg, name, value)
    with cfg.timing.add("import", which="cmd_send"):
        from . import cmd_send

    return go(cmd_send.send, cfg)


# this intermediate function can be mocked by tests that need to build a
# Config object
def go(f, cfg):
    # note: react() does not return
    return react(_dispatch_command, (cfg, lambda: f(cfg)))


# wormhole receive (or "wormhole rx")
@wormhole.command()
@CommonArgs
@TorArgs
@click.option(
    "--only-text",
    "-t",
    is_flag=True,
    help="refuse file transfers, only accept text transfers",
)
@click.option(
    "--accept-file",
    is_flag=True,
    envvar="WORMHOLE_ACCEPT_FILE",
    show_envvar=True,
    help="accept file transfer without asking for confirmation",
)
@click.option(
    "--output-file",
    "-o",
    metavar="FILENAME|DIRNAME",
    help=("The file or directory to create, overriding the name suggested"
          " by the sender. If the directory exists already, the incoming transfer"
          " will be put inside the specified directory."
    ),
)
@click.option(
    "--allocate",
    "-a",
    is_flag=True,
    help="Allocate a fresh code (do not prompt for one)",
)
# --debug-state might be better at the top-level but Click can't parse
# an option like "--debug-state <optional-value>" if there's a subcommand name next
@click.option(
    "--debug-state",
    is_flag=False,
    flag_value="B,N,M,S,O,K,SK,R,RC,L,C,T",
    default=None,
    metavar="MACHINES",
    help=(
        "Debug state-machine transitions. "
        "Possible machines to debug are accepted as a comma-separated list "
        "and the default is all of them. Valid machines are "
        "any of: B,N,M,S,O,K,SK,R,RC,L,C,T"
    )
)
@click.argument(
    "code",
    nargs=-1,
    default=None,
    #    help=("The magic-wormhole code, from the sender. If omitted, the"
    #          " program will ask for it, using tab-completion."),
)
@click.pass_obj
def receive(cfg, code, **kwargs):
    """
    Receive a text message, file, or directory (from 'wormhole send')
    """
    if code:
        if kwargs["allocate"]:
            raise click.UsageError("Cannot specify a code when using --allocate")
    if kwargs["code_length"] and not kwargs["allocate"]:
        raise click.UsageError("Must use --allocate with --code-length")

    kwargs["code_length"] = 2 if kwargs["code_length"] is None else int(kwargs["code_length"])

    for name, value in kwargs.items():
        setattr(cfg, name, value)
    with cfg.timing.add("import", which="cmd_receive"):
        from . import cmd_receive
    if len(code) == 1:
        cfg.code = code[0]
    elif len(code) > 1:
        print("Pass either no code or just one code; you passed"
              " {}: {}".format(len(code), ', '.join(code)))
        raise SystemExit(1)
    else:
        cfg.code = None

    return go(cmd_receive.receive, cfg)


@wormhole.group()
def ssh():
    """
    Facilitate sending/receiving SSH public keys
    """


@ssh.command(name="invite")
@click.option(
    "-c",
    "--code-length",
    default=2,
    metavar="NUMWORDS",
    help="length of code (in bytes/words)",
)
@click.option(
    "--user",
    "-u",
    default=None,
    metavar="USER",
    help="Add to USER's ~/.ssh/authorized_keys",
)
@TorArgs
@click.pass_context
def ssh_invite(ctx, code_length, user, **kwargs):
    """
    Add a public-key to a ~/.ssh/authorized_keys file
    """
    for name, value in kwargs.items():
        setattr(ctx.obj, name, value)
    from . import cmd_ssh
    ctx.obj.code_length = code_length
    ctx.obj.ssh_user = user
    return go(cmd_ssh.invite, ctx.obj)


@ssh.command(name="accept")
@click.argument(
    "code",
    nargs=1,
    required=True,
)
@click.option(
    "--key-file",
    "-F",
    default=None,
    type=click.Path(exists=True),
)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Skip confirmation prompt to send key",
)
@TorArgs
@click.pass_obj
def ssh_accept(cfg, code, key_file, yes, **kwargs):
    """
    Send your SSH public-key

    In response to a 'wormhole ssh invite' this will send public-key
    you specify (if there's only one in ~/.ssh/* that will be sent).
    """

    for name, value in kwargs.items():
        setattr(cfg, name, value)
    from . import cmd_ssh
    kind, keyid, pubkey = cmd_ssh.find_public_key(key_file)
    print(f"Sending public key type='{kind}' keyid='{keyid}'")
    if yes is not True:
        click.confirm(
            f"Really send public key '{keyid}' ?", abort=True)
    cfg.public_key = (kind, keyid, pubkey)
    cfg.code = code

    return go(cmd_ssh.accept, cfg)



from attr import attrib, attrs
from attr.validators import instance_of
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces


@attrs
@implementer(_interfaces.IMailbox)
class Mailbox:
    _side = attrib(validator=instance_of(str))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._mailbox = None
        self._pending_outbound = {}
        self._processed = set()

    def wire(self, nameplate, rendezvous_connector, ordering, terminator):
        self._N = _interfaces.INameplate(nameplate)
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._O = _interfaces.IOrder(ordering)
        self._T = _interfaces.ITerminator(terminator)

    # all -A states: not connected
    # all -B states: yes connected
    # B states serialize as A, so they deserialize as unconnected

    # S0: know nothing
    @m.state(initial=True)
    def S0A(self):
        pass  # pragma: no cover

    @m.state()
    def S0B(self):
        pass  # pragma: no cover

    # S1: mailbox known, not opened
    @m.state()
    def S1A(self):
        pass  # pragma: no cover

    # S2: mailbox known, opened
    # We've definitely tried to open the mailbox at least once, but it must
    # be re-opened with each connection, because open() is also subscribe()
    @m.state()
    def S2A(self):
        pass  # pragma: no cover

    @m.state()
    def S2B(self):
        pass  # pragma: no cover

    # S3: closing
    @m.state()
    def S3A(self):
        pass  # pragma: no cover

    @m.state()
    def S3B(self):
        pass  # pragma: no cover

    # S4: closed. We no longer care whether we're connected or not
    # @m.state()
    # def S4A(self): pass
    # @m.state()
    # def S4B(self): pass
    @m.state(terminal=True)
    def S4(self):
        pass  # pragma: no cover

    S4A = S4
    S4B = S4

    # from Terminator
    @m.input()
    def close(self, mood):
        pass

    # from Nameplate
    @m.input()
    def got_mailbox(self, mailbox):
        pass

    # from RendezvousConnector
    @m.input()
    def connected(self):
        pass

    @m.input()
    def lost(self):
        pass

    def rx_message(self, side, phase, body):
        assert isinstance(side, str), type(side)
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        if side == self._side:
            self.rx_message_ours(phase, body)
        else:
            self.rx_message_theirs(side, phase, body)

    @m.input()
    def rx_message_ours(self, phase, body):
        pass

    @m.input()
    def rx_message_theirs(self, side, phase, body):
        pass

    @m.input()
    def rx_closed(self):
        pass

    # from Send or Key
    @m.input()
    def add_message(self, phase, body):
        pass

    @m.output()
    def record_mailbox(self, mailbox):
        self._mailbox = mailbox

    @m.output()
    def RC_tx_open(self):
        assert self._mailbox
        self._RC.tx_open(self._mailbox)

    @m.output()
    def queue(self, phase, body):
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), (type(body), phase, body)
        self._pending_outbound[phase] = body

    @m.output()
    def record_mailbox_and_RC_tx_open_and_drain(self, mailbox):
        self._mailbox = mailbox
        self._RC.tx_open(mailbox)
        self._drain()

    @m.output()
    def drain(self):
        self._drain()

    def _drain(self):
        for phase, body in self._pending_outbound.items():
            self._RC.tx_add(phase, body)

    @m.output()
    def RC_tx_add(self, phase, body):
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        self._RC.tx_add(phase, body)

    @m.output()
    def N_release_and_accept(self, side, phase, body):
        self._N.release()
        if phase not in self._processed:
            self._processed.add(phase)
            self._O.got_message(side, phase, body)

    @m.output()
    def RC_tx_close(self):
        assert self._mood
        self._RC_tx_close()

    def _RC_tx_close(self):
        self._RC.tx_close(self._mailbox, self._mood)

    @m.output()
    def dequeue(self, phase, body):
        self._pending_outbound.pop(phase, None)

    @m.output()
    def record_mood(self, mood):
        self._mood = mood

    @m.output()
    def record_mood_and_RC_tx_close(self, mood):
        self._mood = mood
        self._RC_tx_close()

    @m.output()
    def ignore_mood_and_T_mailbox_done(self, mood):
        self._T.mailbox_done()

    @m.output()
    def T_mailbox_done(self):
        self._T.mailbox_done()

    S0A.upon(connected, enter=S0B, outputs=[])
    S0A.upon(got_mailbox, enter=S1A, outputs=[record_mailbox])
    S0A.upon(add_message, enter=S0A, outputs=[queue])
    S0A.upon(close, enter=S4A, outputs=[ignore_mood_and_T_mailbox_done])
    S0B.upon(lost, enter=S0A, outputs=[])
    S0B.upon(add_message, enter=S0B, outputs=[queue])
    S0B.upon(close, enter=S4B, outputs=[ignore_mood_and_T_mailbox_done])
    S0B.upon(
        got_mailbox,
        enter=S2B,
        outputs=[record_mailbox_and_RC_tx_open_and_drain])

    S1A.upon(connected, enter=S2B, outputs=[RC_tx_open, drain])
    S1A.upon(add_message, enter=S1A, outputs=[queue])
    S1A.upon(close, enter=S4A, outputs=[ignore_mood_and_T_mailbox_done])

    S2A.upon(connected, enter=S2B, outputs=[RC_tx_open, drain])
    S2A.upon(add_message, enter=S2A, outputs=[queue])
    S2A.upon(close, enter=S3A, outputs=[record_mood])
    S2B.upon(lost, enter=S2A, outputs=[])
    S2B.upon(add_message, enter=S2B, outputs=[queue, RC_tx_add])
    S2B.upon(rx_message_theirs, enter=S2B, outputs=[N_release_and_accept])
    S2B.upon(rx_message_ours, enter=S2B, outputs=[dequeue])
    S2B.upon(close, enter=S3B, outputs=[record_mood_and_RC_tx_close])

    S3A.upon(connected, enter=S3B, outputs=[RC_tx_close])
    S3B.upon(lost, enter=S3A, outputs=[])
    S3B.upon(rx_closed, enter=S4B, outputs=[T_mailbox_done])
    S3B.upon(add_message, enter=S3B, outputs=[])
    S3B.upon(rx_message_theirs, enter=S3B, outputs=[])
    S3B.upon(rx_message_ours, enter=S3B, outputs=[])
    S3B.upon(close, enter=S3B, outputs=[])

    S4A.upon(connected, enter=S4B, outputs=[])
    S4B.upon(lost, enter=S4A, outputs=[])
    S4.upon(add_message, enter=S4, outputs=[])
    S4.upon(rx_message_theirs, enter=S4, outputs=[])
    S4.upon(rx_message_ours, enter=S4, outputs=[])
    S4.upon(close, enter=S4, outputs=[])



import os
import sys

from unittest import mock

from ..cli.public_relay import RENDEZVOUS_RELAY, TRANSIT_RELAY
from .common import config

def test_send_baseline():
    cfg = config("send", "--text", "hi")
    assert cfg.what is None
    assert cfg.code is None
    assert cfg.code_length == 2
    assert cfg.dump_timing is None
    assert not cfg.hide_progress
    assert cfg.listen
    assert cfg.appid is None
    assert cfg.relay_url == RENDEZVOUS_RELAY
    assert cfg.transit_helper == TRANSIT_RELAY
    assert cfg.text == "hi"
    assert not cfg.tor
    assert not cfg.verify
    assert not cfg.zeromode

def test_send_appid():
    cfg = config("--appid", "xyz", "send", "--text", "hi")
    assert cfg.appid == "xyz"
    cfg = config("--appid=xyz", "send", "--text", "hi")
    assert cfg.appid == "xyz"

def test_send_file():
    cfg = config("send", "fn")
    assert cfg.what == "fn"
    assert cfg.text is None

def test_send_text():
    cfg = config("send", "--text", "hi")
    assert cfg.what is None
    assert cfg.text == "hi"

def test_send_nolisten():
    cfg = config("send", "--no-listen", "fn")
    assert not cfg.listen

def test_send_code():
    cfg = config("send", "--code", "1-abc", "fn")
    assert cfg.code == "1-abc"

def test_send_code_length():
    cfg = config("send", "-c", "3", "fn")
    assert cfg.code_length == 3

def test_send_dump_timing():
    cfg = config("--dump-timing", "tx.json", "send", "fn")
    assert cfg.dump_timing == "tx.json"

def test_send_hide_progress():
    cfg = config("send", "--hide-progress", "fn")
    assert cfg.hide_progress

def test_send_tor():
    cfg = config("send", "--tor", "fn")
    assert cfg.tor

def test_send_verify():
    cfg = config("send", "--verify", "fn")
    assert cfg.verify

def test_send_zeromode():
    cfg = config("send", "-0", "fn")
    assert cfg.zeromode

def test_send_relay_env_var():
    relay_url = str(mock.sentinel.relay_url)
    with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
        cfg = config("send")
    assert cfg.relay_url == relay_url

    # Make sure cmd line option overrides environment variable
    relay_url_2 = str(mock.sentinel.relay_url_2)
    with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
        cfg = config("--relay-url", relay_url_2, "send")
    assert cfg.relay_url == relay_url_2

def test_send_transit_env_var():
    transit_url = str(mock.sentinel.transit_url)
    with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
        cfg = config("send")
    assert cfg.transit_helper == transit_url

    # Make sure cmd line option overrides environment variable
    transit_url_2 = str(mock.sentinel.transit_url_2)
    with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
        cfg = config("--transit-helper", transit_url_2, "send")
    assert cfg.transit_helper == transit_url_2


def test_receive_baseline():
    cfg = config("receive")
    assert not cfg.accept_file
    assert cfg.code is None
    assert cfg.code_length == 2
    assert cfg.dump_timing is None
    assert not cfg.hide_progress
    assert cfg.listen
    assert not cfg.only_text
    assert cfg.output_file is None
    assert cfg.appid is None
    assert cfg.relay_url == RENDEZVOUS_RELAY
    assert cfg.transit_helper == TRANSIT_RELAY
    assert not cfg.tor
    assert not cfg.verify
    assert not cfg.zeromode

def test_receive_appid():
    cfg = config("--appid", "xyz", "receive")
    assert cfg.appid == "xyz"
    cfg = config("--appid=xyz", "receive")
    assert cfg.appid == "xyz"

def test_receive_nolisten():
    cfg = config("receive", "--no-listen")
    assert not cfg.listen

def test_receive_code():
    cfg = config("receive", "1-abc")
    assert cfg.code == "1-abc"

def test_receive_code_length():
    cfg = config("receive", "-c", "3", "--allocate")
    assert cfg.code_length == 3

def test_receive_dump_timing():
    cfg = config("--dump-timing", "tx.json", "receive")
    assert cfg.dump_timing == "tx.json"

def test_receive_hide_progress():
    cfg = config("receive", "--hide-progress")
    assert cfg.hide_progress

def test_receive_tor():
    cfg = config("receive", "--tor")
    assert cfg.tor

def test_receive_verify():
    cfg = config("receive", "--verify")
    assert cfg.verify

def test_receive_zeromode():
    cfg = config("receive", "-0")
    assert cfg.zeromode

def test_receive_only_text():
    cfg = config("receive", "-t")
    assert cfg.only_text

def test_receive_accept_file():
    cfg = config("receive", "--accept-file")
    assert cfg.accept_file

def test_receive_output_file():
    cfg = config("receive", "--output-file", "fn")
    assert cfg.output_file == "fn"

def test_receive_relay_env_var():
    relay_url = str(mock.sentinel.relay_url)
    with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
        cfg = config("receive")
    assert cfg.relay_url == relay_url

    # Make sure cmd line option overrides environment variable
    relay_url_2 = str(mock.sentinel.relay_url_2)
    with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
        cfg = config("--relay-url", relay_url_2, "receive")
    assert cfg.relay_url == relay_url_2

def test_receive_transit_env_var():
    transit_url = str(mock.sentinel.transit_url)
    with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
        cfg = config("receive")
    assert cfg.transit_helper == transit_url

    # Make sure cmd line option overrides environment variable
    transit_url_2 = str(mock.sentinel.transit_url_2)
    with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
        cfg = config("--transit-helper", transit_url_2, "receive")
    assert cfg.transit_helper == transit_url_2

def test_receive_accept_file_env_var():
    with mock.patch.dict(os.environ, WORMHOLE_ACCEPT_FILE="true"):
        cfg = config("receive")
    assert cfg.accept_file


def test_receive_send():
    cfg = config("send")
    assert cfg.stdout == sys.stdout

def test_receive_receive():
    cfg = config("receive")
    assert cfg.stdout == sys.stdout



# no unicode_literals until twisted update
from attrs import define
from click.testing import CliRunner
from twisted.application import internet, service
from twisted.internet import defer, endpoints, reactor, protocol
from twisted.python import log

from unittest import mock
from wormhole_mailbox_server.database import create_channel_db, create_usage_db
from wormhole_mailbox_server.server import make_server, Server
from wormhole_mailbox_server.web import make_web_server, PrivacyEnhancedSite
from wormhole_transit_relay.transit_server import Transit, TransitConnection
from wormhole_transit_relay.usage import create_usage_tracker

import sqlite3

from ..cli import cli
from ..transit import allocate_tcp_port


class MyInternetService(service.Service):
    # like StreamServerEndpointService, but you can retrieve the port
    def __init__(self, endpoint, factory):
        self.endpoint = endpoint
        self.factory = factory
        self._port_d = defer.Deferred()
        self._lp = None

    def startService(self):
        super().startService()
        d = self.endpoint.listen(self.factory)

        def good(lp):
            self._lp = lp
            self._port_d.callback(lp.getHost().port)

        def bad(f):
            log.err(f)
            self._port_d.errback(f)

        d.addCallbacks(good, bad)

    async def stopService(self):
        if self._lp:
            await self._lp.stopListening()

    def getPort(self):  # only call once!
        return self._port_d

@define
class Mailbox:
    channel_db: sqlite3.Connection
    usage_db: sqlite3.Connection
    rendezvous: Server
    web: PrivacyEnhancedSite
    url: str
    service: internet.StreamServerEndpointService
    port: object  # IPort ?
    site: object  # ??


async def setup_mailbox(reactor, advertise_version=None, error=None):
    """
    Set up an in-memory Mailbox server.

    If `advertise_version` is not `None`, we advertise it
    If `error` is not `None` we include `error` in the Welcome

    NOTE: Caller is responsible for starting and stopping the service

    :returns: two-tuple of (relay-url, IService instance).
    """
    db = create_channel_db(":memory:")
    usage_db = create_usage_db(":memory:")
    rendezvous = make_server(
        db,
        usage_db=usage_db,
        advertise_version=advertise_version,
        signal_error=error,
    )
    ep = endpoints.TCP4ServerEndpoint(reactor, 0, interface="127.0.0.1")
    site = make_web_server(rendezvous, log_requests=False)
    port = await ep.listen(site)
    service = internet.StreamServerEndpointService(ep, site)
    relay_url = f"ws://127.0.0.1:{port._realPortNumber}/v1"  # XXX private API
    return Mailbox(db, usage_db, rendezvous, site, relay_url, service, port, site)


def setup_transit_relay(reactor):
    transitport = allocate_tcp_port()
    endpoint = f"tcp:{transitport}:interface=127.0.0.1"
    client_endpoint = f"tcp:127.0.0.1:{transitport}"
    ep = endpoints.serverFromString(reactor, endpoint)
    usage = create_usage_tracker(blur_usage=None, log_file=None, usage_db=None)
    transit_server = protocol.ServerFactory()
    transit_server.protocol = TransitConnection
    transit_server.log_requests = False
    transit_server.transit = Transit(usage, reactor.seconds)
    service = internet.StreamServerEndpointService(ep, transit_server)
    return client_endpoint, service


# XXX missing from ServerBase replacement fixtures:

        # Unit tests that spawn a (blocking) client in a thread might still
        # have threads running at this point, if one is stuck waiting for a
        # message from a companion which has exited with an error. Our
        # relay's .stopService() drops all connections, which ought to
        # encourage those threads to terminate soon. If they don't, print a
        # warning to ease debugging.

        # XXX FIXME there's something in _noclobber test that's not
        # waiting for a close, I think -- was pretty relieably getting
        # unclean-reactor, but adding a slight pause here stops it...

"""
        tp = reactor.getThreadPool()
        if not tp.working:
            await self.sp.stopService()
            await task.deferLater(reactor, 0.1, lambda: None)
            return None
        # disconnect all callers
        d = defer.maybeDeferred(self.sp.stopService)
        d.addBoth(lambda _: self._transit_server.stopFactory())
        # wait a second, then check to see if it worked
        await task.deferLater(reactor, 1.0, lambda: None)
        if len(tp.working):
            log.msg("wormhole.test.common.ServerBase.tearDown:"
                    " I was unable to convince all threads to exit.")
            tp.dumpStats()
            print("tearDown warning: threads are still active")
            print("This test will probably hang until one of the"
                  " clients gives up of their own accord.")
        else:
            log.msg("wormhole.test.common.ServerBase.tearDown:"
                    " I convinced all threads to exit.")
        await d
"""


def config(*argv):
    r = CliRunner()
    with mock.patch("wormhole.cli.cli.go") as go:
        res = r.invoke(cli.wormhole, argv, catch_exceptions=False)
        if res.exit_code != 0:
            print(res.exit_code)
            print(res.output)
            print(res)
            assert 0
        cfg = go.call_args[0][1]
    return cfg


async def poll_until(predicate):
    # return a Deferred that won't fire until the predicate is True
    while not predicate():
        d = defer.Deferred()
        reactor.callLater(0.001, d.callback, None)
        await d



import io
import re

from twisted.internet.defer import gatherResults
from twisted.internet.error import ConnectionRefusedError

from unittest import mock
from pytest_twisted import ensureDeferred

from .. import _rendezvous, wormhole
from ..errors import (KeyFormatError, LonelyError, NoKeyError,
                      OnlyOneCodeError, ServerConnectionError, WormholeClosed,
                      WrongPasswordError)
from ..eventual import EventualQueue
from ..transit import allocate_tcp_port
from .common import poll_until
import pytest

APPID = "appid"

# event orderings to exercise:
#
# * normal sender: set_code, send_phase1, connected, claimed, learn_msg2,
#   learn_phase1
# * normal receiver (argv[2]=code): set_code, connected, learn_msg1,
#   learn_phase1, send_phase1,
# * normal receiver (readline): connected, input_code
# *
# * set_code, then connected
# * connected, receive_pake, send_phase, set_code


class Delegate:
    def __init__(self):
        self.welcome = None
        self.code = None
        self.key = None
        self.verifier = None
        self.versions = None
        self.messages = []
        self.closed = None

    def wormhole_got_welcome(self, welcome):
        self.welcome = welcome

    def wormhole_got_code(self, code):
        self.code = code

    def wormhole_got_unverified_key(self, key):
        self.key = key

    def wormhole_got_verifier(self, verifier):
        self.verifier = verifier

    def wormhole_got_versions(self, versions):
        self.versions = versions

    def wormhole_got_message(self, data):
        self.messages.append(data)

    def wormhole_closed(self, result):
        self.closed = result


@ensureDeferred
async def test_delegated(reactor, mailbox):
    dg = Delegate()
    w1 = wormhole.create(APPID, mailbox.url, reactor, delegate=dg)
    # w1.debug_set_trace("W1")
    with pytest.raises(NoKeyError):
        w1.derive_key("purpose", 12)
    w1.set_code("1-abc")
    assert dg.code == "1-abc"
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w2.set_code(dg.code)
    await poll_until(lambda: dg.key is not None)
    await poll_until(lambda: dg.verifier is not None)
    await poll_until(lambda: dg.versions is not None)

    w1.send_message(b"ping")
    got = await w2.get_message()
    assert got == b"ping"
    w2.send_message(b"pong")
    await poll_until(lambda: dg.messages)
    assert dg.messages[0] == b"pong"

    key1 = w1.derive_key("purpose", 16)
    assert len(key1) == 16
    assert type(key1) is bytes
    with pytest.raises(TypeError):
        w1.derive_key(b"not unicode", 16)
    with pytest.raises(TypeError):
        w1.derive_key(12345, 16)

    w1.close()
    await w2.close()


@ensureDeferred
async def test_delegate_allocate_code(reactor, mailbox):
    dg = Delegate()
    w1 = wormhole.create(APPID, mailbox.url, reactor, delegate=dg)
    w1.allocate_code()
    await poll_until(lambda: dg.code is not None)
    w1.close()


@ensureDeferred
async def test_delegate_input_code(reactor, mailbox):
    dg = Delegate()
    w1 = wormhole.create(APPID, mailbox.url, reactor, delegate=dg)
    h = w1.input_code()
    h.choose_nameplate("123")
    h.choose_words("purple-elephant")
    await poll_until(lambda: dg.code is not None)
    w1.close()


# integration test, with a real server

async def doBoth(d1, d2):
    return await gatherResults([d1, d2], True)


@ensureDeferred
async def test_allocate_default(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w1.allocate_code()
    code = await w1.get_code()
    mo = re.search(r"^\d+-\w+-\w+$", code)
    assert mo, code
    # w.close() fails because we closed before connecting
    with pytest.raises(LonelyError):
        await w1.close()


@ensureDeferred
async def test_allocate_more_words(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w1.allocate_code(3)
    code = await w1.get_code()
    mo = re.search(r"^\d+-\w+-\w+-\w+$", code)
    assert mo, code
    with pytest.raises(LonelyError):
        await w1.close()


@ensureDeferred
async def test_basic(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    # w1.debug_set_trace("W1")
    with pytest.raises(NoKeyError):
        w1.derive_key("purpose", 12)

    w2 = wormhole.create(APPID, mailbox.url, reactor)
    # w2.debug_set_trace("  W2")
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)

    await w1.get_unverified_key()
    await w2.get_unverified_key()

    key1 = w1.derive_key("purpose", 16)
    assert len(key1) == 16
    assert type(key1) is bytes
    with pytest.raises(TypeError):
        w1.derive_key(b"not unicode", 16)
    with pytest.raises(TypeError):
        w1.derive_key(12345, 16)

    verifier1 = await w1.get_verifier()
    verifier2 = await w2.get_verifier()
    assert verifier1 == verifier2

    versions1 = await w1.get_versions()
    versions2 = await w2.get_versions()
    # app-versions are exercised properly in test_versions, this just
    # tests the defaults
    assert versions1 == {}
    assert versions2 == {}

    w1.send_message(b"data1")
    w2.send_message(b"data2")
    dataX = await w1.get_message()
    dataY = await w2.get_message()
    assert dataX == b"data2"
    assert dataY == b"data1"

    versions1_again = await w1.get_versions()
    assert versions1 == versions1_again

    c1 = await w1.close()
    assert c1 == "happy"
    c2 = await w2.close()
    assert c2 == "happy"


@ensureDeferred
async def test_get_code_early(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    d = w1.get_code()
    w1.set_code("1-abc")
    await eq.flush()
    code = await d
    assert code == "1-abc"
    with pytest.raises(LonelyError):
        await w1.close()


@ensureDeferred
async def test_get_code_late(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w1.set_code("1-abc")
    d = w1.get_code()
    await eq.flush()
    code = await d
    assert code == "1-abc"
    with pytest.raises(LonelyError):
        await w1.close()


@ensureDeferred
async def test_same_message(reactor, mailbox):
    # the two sides use random nonces for their messages, so it's ok for
    # both to try and send the same body: they'll result in distinct
    # encrypted messages
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    w1.send_message(b"data")
    w2.send_message(b"data")
    dataX = await w1.get_message()
    dataY = await w2.get_message()
    assert dataX == b"data"
    assert dataY == b"data"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_interleaved(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    w1.send_message(b"data1")
    dataY = await w2.get_message()
    assert dataY == b"data1"
    d = w1.get_message()
    w2.send_message(b"data2")
    dataX = await d
    assert dataX == b"data2"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_unidirectional(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    w1.send_message(b"data1")
    dataY = await w2.get_message()
    assert dataY == b"data1"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_early(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w1.send_message(b"data1")
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    d = w2.get_message()
    w1.set_code("123-abc-def")
    w2.set_code("123-abc-def")
    dataY = await d
    assert dataY == b"data1"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_fixed_code(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.set_code("123-purple-elephant")
    w2.set_code("123-purple-elephant")
    w1.send_message(b"data1"), w2.send_message(b"data2")
    dl = await doBoth(w1.get_message(), w2.get_message())
    (dataX, dataY) = dl
    assert dataX == b"data2"
    assert dataY == b"data1"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_input_code(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.set_code("123-purple-elephant")
    h = w2.input_code()
    h.choose_nameplate("123")
    # Pause to allow some messages to get delivered. Specifically we want
    # to wait until w2 claims the nameplate, opens the mailbox, and
    # receives the PAKE message, to exercise the PAKE-before-CODE path in
    # Key.
    await poll_until(lambda: w2._boss._K._debug_pake_stashed)
    h.choose_words("purple-elephant")

    w1.send_message(b"data1"), w2.send_message(b"data2")
    dl = await doBoth(w1.get_message(), w2.get_message())
    (dataX, dataY) = dl
    assert dataX == b"data2"
    assert dataY == b"data1"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_multiple_messages(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.set_code("123-purple-elephant")
    w2.set_code("123-purple-elephant")
    w1.send_message(b"data1"), w2.send_message(b"data2")
    w1.send_message(b"data3"), w2.send_message(b"data4")
    dl = await doBoth(w1.get_message(), w2.get_message())
    (dataX, dataY) = dl
    assert dataX == b"data2"
    assert dataY == b"data1"
    dl = await doBoth(w1.get_message(), w2.get_message())
    (dataX, dataY) = dl
    assert dataX == b"data4"
    assert dataY == b"data3"
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_closed(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w2 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w1.set_code("123-foo")
    w2.set_code("123-foo")

    # let it connect and become HAPPY
    await w1.get_versions()
    await w2.get_versions()

    await w1.close()
    await w2.close()

    # once closed, all Deferred-awaiting API calls get an prompt error
    with pytest.raises(WormholeClosed):
        await w1.get_welcome()
    with pytest.raises(WormholeClosed) as f:
        await w1.get_code()
    assert f.value.args[0] == "happy"
    with pytest.raises(WormholeClosed):
        await w1.get_unverified_key()
    with pytest.raises(WormholeClosed):
        await w1.get_verifier()
    with pytest.raises(WormholeClosed):
        await w1.get_versions()
    with pytest.raises(WormholeClosed):
        await w1.get_message()


@ensureDeferred
async def test_closed_idle(reactor):
    port = allocate_tcp_port()
    # without a relay server, this won't ever connect
    w1 = wormhole.create(APPID, f"ws://127.0.0.1:{port}/v1", reactor)

    d_welcome = w1.get_welcome()
    assert not d_welcome.called
    d_code = w1.get_code()
    d_key = w1.get_unverified_key()
    d_verifier = w1.get_verifier()
    d_versions = w1.get_versions()
    d_message = w1.get_message()

    with pytest.raises(LonelyError):
        await w1.close()

    with pytest.raises(LonelyError):
        await d_welcome
    with pytest.raises(LonelyError):
        await d_code
    with pytest.raises(LonelyError):
        await d_key
    with pytest.raises(LonelyError):
        await d_verifier
    with pytest.raises(LonelyError):
        await d_versions
    with pytest.raises(LonelyError):
        await d_message


@ensureDeferred
async def test_wrong_password(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w2 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code + "not")
    code2 = await w2.get_code()
    assert code != code2
    # That's enough to allow both sides to discover the mismatch, but
    # only after the confirmation message gets through. API calls that
    # don't wait will appear to work until the mismatched confirmation
    # message arrives.
    w1.send_message(b"should still work")
    w2.send_message(b"should still work")

    key2 = await w2.get_unverified_key()  # should work
    # w2 has just received w1.PAKE, and is about to send w2.VERSION
    key1 = await w1.get_unverified_key()  # should work
    # w1 has just received w2.PAKE, and is about to send w1.VERSION, and
    # then will receive w2.VERSION. When it sees w2.VERSION, it will
    # learn about the WrongPasswordError.
    assert key1 != key2

    # API calls that wait (i.e. get) will errback. We collect all these
    # Deferreds early to exercise the wait-then-fail path
    d1_verified = w1.get_verifier()
    d1_versions = w1.get_versions()
    d1_received = w1.get_message()
    d2_verified = w2.get_verifier()
    d2_versions = w2.get_versions()
    d2_received = w2.get_message()

    # wait for each side to notice the failure
    with pytest.raises(WrongPasswordError):
        await w1.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w2.get_verifier()
    # the rest of the loops should fire within the next tick
    await eq.flush()

    # now all the rest should have fired already
    with pytest.raises(WrongPasswordError):
        await d1_verified
    with pytest.raises(WrongPasswordError):
        await d1_versions
    with pytest.raises(WrongPasswordError):
        await d1_received
    with pytest.raises(WrongPasswordError):
        await d2_verified
    with pytest.raises(WrongPasswordError):
        await d2_versions
    with pytest.raises(WrongPasswordError):
        await d2_received

    # and at this point, with the failure safely noticed by both sides,
    # new get_unverified_key() calls should signal the failure, even
    # before we close

    # any new calls in the error state should immediately fail
    with pytest.raises(WrongPasswordError):
        await w1.get_unverified_key()
    with pytest.raises(WrongPasswordError):
        await w1.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w1.get_versions()
    with pytest.raises(WrongPasswordError):
        await w1.get_message()
    with pytest.raises(WrongPasswordError):
        await w2.get_unverified_key()
    with pytest.raises(WrongPasswordError):
        await w2.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w2.get_versions()
    with pytest.raises(WrongPasswordError):
        await w2.get_message()

    with pytest.raises(WrongPasswordError):
        await w1.close()
    with pytest.raises(WrongPasswordError):
        await w2.close()

    # API calls should still get the error, not WormholeClosed
    with pytest.raises(WrongPasswordError):
        await w1.get_unverified_key()
    with pytest.raises(WrongPasswordError):
        await w1.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w1.get_versions()
    with pytest.raises(WrongPasswordError):
        await w1.get_message()
    with pytest.raises(WrongPasswordError):
        await w2.get_unverified_key()
    with pytest.raises(WrongPasswordError):
        await w2.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w2.get_versions()
    with pytest.raises(WrongPasswordError):
        await w2.get_message()


@ensureDeferred
async def test_wrong_password_with_spaces(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    badcode = "4 oops spaces"
    with pytest.raises(KeyFormatError) as ex:
        w.set_code(badcode)
    expected_msg = f"Code '{badcode}' contains spaces."
    assert expected_msg == str(ex.value)
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_wrong_password_with_leading_space(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    badcode = " 4-oops-space"
    with pytest.raises(KeyFormatError) as ex:
        w.set_code(badcode)
    expected_msg = f"Code '{badcode}' contains spaces."
    assert expected_msg == str(ex.value)
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_wrong_password_with_non_numeric_nameplate(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    badcode = "four-oops-space"
    with pytest.raises(KeyFormatError) as ex:
        w.set_code(badcode)
    expected_msg = "Nameplate 'four' must be numeric, with no spaces."
    assert expected_msg == str(ex.value)
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_welcome(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    wel1 = await w1.get_welcome()  # early: before connection established
    wel2 = await w1.get_welcome()  # late: already received welcome
    assert wel1 == wel2
    assert "current_cli_version" in wel1

    # cause an error, so a later get_welcome will return the error
    w1.set_code("123-foo")
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w2.set_code("123-NOT")
    with pytest.raises(WrongPasswordError):
        await w1.get_verifier()
    with pytest.raises(WrongPasswordError):
        await w1.get_welcome()

    # we have to ensure w2 receives a "bad" message from w1 before
    # the w2.close() assertion below will actually fail
    with pytest.raises(WrongPasswordError):
        await w2.get_verifier()

    with pytest.raises(WrongPasswordError):
        await w1.close()
    with pytest.raises(WrongPasswordError):
        await w2.close()


@ensureDeferred
async def test_verifier(reactor, mailbox):
    eq = EventualQueue(reactor)
    w1 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w2 = wormhole.create(APPID, mailbox.url, reactor, _eventual_queue=eq)
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    v1 = await w1.get_verifier()  # early
    v2 = await w2.get_verifier()
    assert type(v1) is bytes
    assert v1 == v2
    w1.send_message(b"data1")
    w2.send_message(b"data2")
    dataX = await w1.get_message()
    dataY = await w2.get_message()
    assert dataX == b"data2"
    assert dataY == b"data1"

    # calling get_verifier() this late should fire right away
    d = w2.get_verifier()
    await eq.flush()
    v1_late = await d
    assert v1_late == v1

    await w1.close()
    await w2.close()


@ensureDeferred
async def test_versions(reactor, mailbox):
    # there's no API for this yet, but make sure the internals work
    w1 = wormhole.create(
        APPID, mailbox.url, reactor, versions={"w1": 123})
    w2 = wormhole.create(
        APPID, mailbox.url, reactor, versions={"w2": 456})
    w1.allocate_code()
    code = await w1.get_code()
    w2.set_code(code)
    w1_versions = await w2.get_versions()
    assert w1_versions == {"w1": 123}
    w2_versions = await w1.get_versions()
    assert w2_versions == {"w2": 456}
    await w1.close()
    await w2.close()


@ensureDeferred
async def test_rx_dedup(reactor, mailbox):
    # Future clients will handle losing/reestablishing the Rendezvous
    # Server connection by retransmitting messages, which will sometimes
    # cause duplicate messages. Make sure this client can tolerate them.
    # The first place this would fail was when the second copy of the
    # incoming PAKE message was received, which would cause
    # SPAKE2.finish() to be called a second time, which throws an error
    # (which, being somewhat unexpected, caused a hang rather than a
    # clear exception). The Mailbox object is responsible for
    # deduplication, so we must patch the RendezvousConnector to simulate
    # duplicated messages.
    with mock.patch("wormhole._boss.RendezvousConnector", MessageDoubler):
        w1 = wormhole.create(APPID, mailbox.url, reactor)
    w2 = wormhole.create(APPID, mailbox.url, reactor)
    w1.set_code("123-purple-elephant")
    w2.set_code("123-purple-elephant")
    w1.send_message(b"data1"), w2.send_message(b"data2")
    dl = await doBoth(w1.get_message(), w2.get_message())
    (dataX, dataY) = dl
    assert dataX == b"data2"
    assert dataY == b"data1"
    await w1.close()
    await w2.close()


class MessageDoubler(_rendezvous.RendezvousConnector):
    # we could double messages on the sending side, but a future server will
    # strip those duplicates, so to really exercise the receiver, we must
    # double them on the inbound side instead
    # def _msg_send(self, phase, body):
    #     wormhole._Wormhole._msg_send(self, phase, body)
    #     self._ws_send_command("add", phase=phase, body=bytes_to_hexstr(body))
    def _response_handle_message(self, msg):
        _rendezvous.RendezvousConnector._response_handle_message(self, msg)
        _rendezvous.RendezvousConnector._response_handle_message(self, msg)


@ensureDeferred
async def test_derive_key_early(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    # definitely too early
    with pytest.raises(NoKeyError):
        w.derive_key("purpose", 12)
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_multiple_set_code(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    w.set_code("123-purple-elephant")
    # code can only be set once
    with pytest.raises(OnlyOneCodeError):
        w.set_code("123-nope")
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_allocate_and_set_code(reactor, mailbox):
    w = wormhole.create(APPID, mailbox.url, reactor)
    w.allocate_code()
    await w.get_code()
    with pytest.raises(OnlyOneCodeError):
        w.set_code("123-nope")
    with pytest.raises(LonelyError):
        await w.close()


@ensureDeferred
async def test_reconnection_basic(reactor, mailbox):
    w1 = wormhole.create(APPID, mailbox.url, reactor)
    w1_in = []
    w1._boss._RC._debug_record_inbound_f = w1_in.append
    # w1.debug_set_trace("W1")
    w1.allocate_code()
    code = await w1.get_code()
    w1.send_message(b"data1")  # queued until wormhole is established

    # now wait until we've deposited all our messages on the server
    def seen_our_pake():
        for m in w1_in:
            if m["type"] == "message" and m["phase"] == "pake":
                return True
        return False

    await poll_until(seen_our_pake)

    w1_in[:] = []
    # drop the connection
    w1._boss._RC._ws.transport.loseConnection()
    # wait for it to reconnect and redeliver all the messages. The server
    # sends mtype=message messages in random order, but we've only sent
    # one of them, so it's safe to wait for just the PAKE phase.
    await poll_until(seen_our_pake)

    # now let the second side proceed. this simulates the most common
    # case: the server is bounced while the sender is waiting, before the
    # receiver has started

    w2 = wormhole.create(APPID, mailbox.url, reactor)
    # w2.debug_set_trace("  W2")
    w2.set_code(code)

    dataY = await w2.get_message()
    assert dataY == b"data1"

    w2.send_message(b"data2")
    dataX = await w1.get_message()
    assert dataX == b"data2"

    c1 = await w1.close()
    assert c1 == "happy"
    c2 = await w2.close()
    assert c2 == "happy"


@ensureDeferred
async def assertSCEFailure(eq, d, innerType):
    await eq.flush()
    with pytest.raises(ServerConnectionError) as f:
        await d
    inner = f.value.reason
    assert isinstance(inner, innerType)
    return inner


@ensureDeferred
async def test_bad_dns(reactor):
    eq = EventualQueue(reactor)
    # point at a URL that will never connect
    w = wormhole.create(
        APPID, "ws://%%%.example.org:4000/v1", reactor, _eventual_queue=eq)
    # that should have already received an error, when it tried to
    # resolve the bogus DNS name. All API calls will return an error.

    e = await assertSCEFailure(eq, w.get_unverified_key(), ValueError)
    assert isinstance(e, ValueError)
    assert str(e) == "invalid hostname: %%%.example.org"
    await assertSCEFailure(eq, w.get_code(), ValueError)
    await assertSCEFailure(eq, w.get_verifier(), ValueError)
    await assertSCEFailure(eq, w.get_versions(), ValueError)
    await assertSCEFailure(eq, w.get_message(), ValueError)


@ensureDeferred
async def assertSCE(d, innerType):
    with pytest.raises(ServerConnectionError) as f:
        await d
    inner = f.value.reason
    assert isinstance(inner, innerType)
    return inner


@ensureDeferred
async def test_no_connection(reactor):
    # point at a URL that will never connect
    port = allocate_tcp_port()
    w = wormhole.create(APPID, f"ws://127.0.0.1:{port}/v1", reactor)
    # nothing is listening, but it will take a turn to discover that
    d1 = w.get_code()
    d2 = w.get_unverified_key()
    d3 = w.get_verifier()
    d4 = w.get_versions()
    d5 = w.get_message()
    await assertSCE(d1, ConnectionRefusedError)
    await assertSCE(d2, ConnectionRefusedError)
    await assertSCE(d3, ConnectionRefusedError)
    await assertSCE(d4, ConnectionRefusedError)
    await assertSCE(d5, ConnectionRefusedError)


@ensureDeferred
async def test_all_deferreds(reactor):
    # point at a URL that will never connect
    port = allocate_tcp_port()
    w = wormhole.create(APPID, f"ws://127.0.0.1:{port}/v1", reactor)
    # nothing is listening, but it will take a turn to discover that
    w.allocate_code()
    d1 = w.get_code()
    d2 = w.get_unverified_key()
    d3 = w.get_verifier()
    d4 = w.get_versions()
    d5 = w.get_message()
    await assertSCE(d1, ConnectionRefusedError)
    await assertSCE(d2, ConnectionRefusedError)
    await assertSCE(d3, ConnectionRefusedError)
    await assertSCE(d4, ConnectionRefusedError)
    await assertSCE(d5, ConnectionRefusedError)


def test_trace_basic(reactor):
    w1 = wormhole.create(APPID, "ws://localhost:1", reactor)
    stderr = io.StringIO()
    w1.debug_set_trace("W1", file=stderr)
    # if Automat doesn't have the tracing API, then we won't actually
    # exercise the tracing function, so exercise the RendezvousConnector
    # function manually (it isn't a state machine, so it will always wire
    # up the tracer)
    w1._boss._RC._debug("what")

    stderr = io.StringIO()
    out = w1._boss._print_trace("OLD", "IN", "NEW", "C1", "M1", stderr)
    assert stderr.getvalue().splitlines() == \
                     ["C1.M1[OLD].IN -> [NEW]"]
    out("OUT1")
    assert stderr.getvalue().splitlines() == \
                     ["C1.M1[OLD].IN -> [NEW]", " C1.M1.OUT1()"]
    w1._boss._print_trace("", "R.connected", "", "C1", "RC1", stderr)
    assert stderr.getvalue().splitlines() == \
        ["C1.M1[OLD].IN -> [NEW]", " C1.M1.OUT1()", "C1.RC1.R.connected"]


def test_trace_delegated(reactor):
    dg = Delegate()
    w1 = wormhole.create(APPID, "ws://localhost:1", reactor, delegate=dg)
    stderr = io.StringIO()
    w1.debug_set_trace("W1", file=stderr)
    w1._boss._RC._debug("what")



from zope.interface import Interface

# These interfaces are private: we use them as markers to detect
# swapped argument bugs in the various .wire() calls


class IWormhole(Interface):
    """Internal: this contains the methods invoked 'from below'."""

    def got_welcome(welcome):
        pass

    def got_code(code):
        pass

    def got_key(key):
        pass

    def got_verifier(verifier):
        pass

    def got_versions(versions):
        pass

    def received(plaintext):
        pass

    def closed(result):
        pass


class IWormholeDelegate(Interface):
    """
    Must be implemented by the object passed as "delegate=" to wormhole.create

    These methods will be called at the appropriate time by the
    _DelegatedWormhole instance. They will be called from the main
    thread which is running the reactor. Methods MUST NOT block for
    "too long"; use deferToThread() or similar mechanisms if blocking
    code needs to run.
    """

    def wormhole_got_welcome(welcome: dict) -> None:
        """
        We have connected to the server and received its "welcome"
        message, which is an arbitrary `dict`. It may contain a
        `"motd"` attribute which -- if present -- is intended to be
        shown to the user.

        Other attributes may be present.
        """

    def wormhole_got_code(code: str) -> None:
        """
        We have a secret code. This could be because the wormhole was told
        to allocate one, or because one was set.
        """

    def wormhole_got_unverified_key(key: bytes) -> None:  # XXX definitely bytes?
        """
        We have received the symmetric, secret key. It is "unverified"
        because we have not yet successfully exchanged a message
        confirming that the other side has the same key; the first
        such message is the "versions" message.

        It is also possible for the users to confirm that the same key
        is being used by comparing the "verifier" strings (which are a
        hash of the key).
        """

    def wormhole_got_verifier(verifier: bytes) -> None:  # XXX bytes, or str?
        """
        A tagged hash of the secret key. If both sides compare their
        verifier (usually out-of-band) they can be assured that they
        both have the same secret key and are thus not subject to a
        MitM
        """

    def wormhole_got_versions(versions: dict) -> None:
        """
        We have received the "versions" message from our peer.

        This confirms that we've got a shared secret key (or else this
        message wouldn't decrypt).

        It is also used to exchange arbitrary information, intended
        for applications to learn what the peer supports.
        The `dict` may contain any JSON-able information.
        """

    def wormhole_got_message(plaintext: bytes) -> None:  # XXX bytes, or str?
        """
        We have received a message from the other side during normal
        operations (i.e. after "versions" has been exchanged).

        These are application-specific messages.
        """

    def wormhole_closed(result: str) -> None:
        """
        The wormhole session has been closed. The `result` string is one
        of the following:

        - `"empty"`: we didn't get very far
        - `"unwelcome"`: XXX didn't get welcome? permissions failed?
        - `"errory"`: something went wrong
        - `"scary"`: key-confirmation failed
        - `"lonely"`: our peer never showed up
        - `"happy"`: everything worked as intended
        """


class IBoss(Interface):
    pass


class INameplate(Interface):
    pass


class IMailbox(Interface):
    pass


class ISend(Interface):
    pass


class IOrder(Interface):
    pass


class IKey(Interface):
    pass


class IReceive(Interface):
    pass


class IRendezvousConnector(Interface):
    pass


class ILister(Interface):
    pass


class ICode(Interface):
    pass


class IInput(Interface):
    pass


class IAllocator(Interface):
    pass


class ITerminator(Interface):
    pass


class ITiming(Interface):
    pass


class ITorManager(Interface):
    pass


class IWordlist(Interface):
    def choose_words(length):
        """Randomly select LENGTH words, join them with hyphens, return the
        result."""

    def get_completions(prefix):
        """Return a list of all suffixes that could complete the given
        prefix."""


# These interfaces are public, and are re-exported by __init__.py


class IDeferredWormhole(Interface):
    def get_welcome():
        """
        Wait for the 'welcome message' dictionary, sent by the server upon
        first connection.

        :rtype: ``Deferred[dict]``
        :return: the welcome dictionary, when it arrives from the server
        """

    def allocate_code(code_length=2):
        """
        Ask the wormhole to allocate a nameplate and generate a random code.

        When the code is ready, any Deferreds returned by ``get_code()`` will
        be fired.  Only one of generate_code/set_code/input_code may be used.

        :param int code_length: the number of random words to use.  More
            words means the code is harder to guess.  Defaults to 2.

        :return: None

            ~mod.class
        """

    def set_code(code):
        """
        Tell the wormhole to use a specific code, either copied from a
        wormhole that used ``allocate_code``, or created out-of-band by
        humans (and given to ``set_code`` on both wormholes).

        Any Deferreds returned by ``get_code()`` will be fired as soon as
        this is called.  Only one of generate_code/set_code/input_code may be
        used.

        :return: None
        """

    def input_code():
        """
        Ask the wormhole to perform interactive entry of the code, with
        completion on the nameplate and/or words.

        This does not actually interact with the user, but instead returns a
        'code-entry helper' object.  The application is responsible for doing
        the IO: the helper is used to get completion lists and to submit the
        finished code.  See ``input_with_completion`` for a wrapper function
        that uses ``readline`` to do CLI-style input completion.

        Any Deferreds returned by ``get_code()`` will be fired when the final
        code is submitted to the helper.  Only one of
        generate_code/set_code/input_code may be used.

        :return: a code-entry helper instance
        :rtype: IHelper
        """

    def get_code():
        """
        Wait for the wormhole code to be established, then return the code.
        This is really only useful on the initiating side, which needs to
        deliver the code to the user (so the user can dictate it to the other
        user, who can deliver it to their application with ``set_code`` or
        ``input_code``).  On the receiving side, merely submitting the code
        is sufficient.

        The wormhole code is always unicode (so ``str`` on py3, ``unicode``
        on py2).

        For ``allocate_code``, this must wait for the server to allocate a
        nameplate.  For ``input_code``, it waits for the final code to be
        submitted to the helper.  For ``set_code``, it fires immediately.

        :return: the wormhole code
        :rtype: ``Deferred[str]``
        """

    def get_unverified_key():
        """
        Wait for key-exchange to occur, then return the raw unverified SPAKE2
        key.  When this fires, we have not seen any evidence that anyone else
        shares this key (nor have we seen evidence of a failed attack: e.g. a
        payload encrypted with a different key).

        This is only useful for testing, and for noticing a significant delay
        between the key-agreement message and the subsequent key-verification
        ("versions") message.

        :return: the raw unverified SPAKE2 key
        :rtype: ``Deferred[bytes]``
        """

    def get_verifier():
        """
        Wait for key verification to occur, then return the verifier string.
        When this fires, we have seen at least one validly-encrypted message
        from our peer, indicating that we have established a shared secret
        key with some party who knows (or correctly guessed) the wormhole
        code.

        The verifier string (bytes) can be displayed to the user (perhaps as
        hex), who can manually compare it with the peer's verifier, to obtain
        more confidence in the secrecy of the established key.

        If we receive an invalid encrypted message (such as what would happen
        if an attacker tried and failed to guess the wormhole code), this
        will instead errback with a ``WrongPasswordError``.

        :return: the verifier string, after a valid encrypted message has
                 arrived
        :rtype: ``Deferred[bytes]``
        """

    def get_versions():
        """
        Wait for a valid VERSION message to arrive, then return the peer's
        "versions" dictionary.  This dictionary comes from the ``versions=``
        argument to the peer's ``wormhole()`` constructor, and is meant to
        assist with capability-negotiation between the two peers.  In
        particular, the ``versions`` dictionary is delivered before either
        side has called ``send_message()``, so it can influence the first
        message sent to a peer that is too old to use that first message for
        negotiation purposes.

        If we receive any invalid encrypted message (such as what would
        happen if an attacker tried and failed to guess the wormhole code),
        this will instead errback with a ``WrongPasswordError``.

        :return: the versions dictionary
        :rtype: ``Deferred[dict]``
        """

    def derive_key(purpose, length):
        """
        Derive a purpose-specific key.

        This combines the master SPAKE2 key with the given purpose string and
        deterministically derives a new key of the requested length.  Any two
        connected Wormhole objects which call ``derive_key`` with the same
        purpose and length will get the same key.  This can be used to
        encrypt or sign other messages, or exchanged for verification
        purposes.  The master key will remain secret even if you reveal a
        derivative key.

        This must be called after the key has been established, so after any
        of
        ``get_unverified_key()/get_verifier()/get_versions()/get_message()``
        have fired.  ``derive_key()`` returns immediately, rather than
        returning a ``Deferred``.

        :return: a derivative key, of the requested length
        :rtype: ``bytes``
        """

    def send_message(msg):
        """
        Send a message to the connected peer.

        This accepts a bytestring, and queues it for encryption and delivery
        to the other side, where it will eventually appear in
        ``get_message()``.  Messages are delivered in-order, and complete
        (the Wormhole is a record-pipe, not a byte-pipe).

        This can be called at any time, even before setting the wormhole
        code.  The message will be queued for delivery after the master key
        is established.

        :return: None
        """

    def get_message():
        """
        Wait for, and return, the next message.

        This returns a Deferred that will fire when the next (sequential)
        application message has been received and successfully decrypted.
        Messages will be delivered in-order and intact (the Wormhole is a
        record-pipe, not a byte-pipe).

        This can be called at any time, even before setting the wormhole
        code.  The Deferred will not fire until key-negotiation has completed
        and a validly-encrypted message has arrived.

        If we receive any invalid encrypted message (such as what would
        happen if an attacker tried and failed to guess the wormhole code),
        this will instead errback with a ``WrongPasswordError``.

        :return: the next decrypted message
        :rtype: ``Deferred[bytes]``
        """

    def close():
        """
        Close the wormhole.

        This frees all resources associated with the wormhole (including
        server-side queues and any established network connections).

        For operational purposes, it informs the server that the wormhole
        closed "happy".  Less-happy moods may be reported if the connection
        closed due to a ``WrongPasswordError`` or because of a timeout.

        ``close()`` returns a Deferred, which fires after the server has been
        informed and the sockets have been shut down.  One-shot applications
        should delay shutdown until this Deferred has fired, to increase the
        chances that server resources will be freed.  Long-running
        applications can probably ignore the Deferred, as they'll probably
        remain running long enough to allow the shutdown to complete.

        The Deferred will errback if the wormhole had problems, like a
        ``WrongPasswordError``.

        :return: a Deferred that fires when shutdown is complete
        :rtype: ``Deferred``
        """


class IInputHelper(Interface):
    def refresh_nameplates():
        """
        Refresh the nameplates list.

        This asks the server for the set of currently-active nameplates
        (either from calls to ``allocate_code()`` or referenced by active
        wormhole clients).  It updates the set available to
        ``get_nameplate_completions()``.

        :return: None
        """

    def get_nameplate_completions(prefix):
        """
        Return a list of nameplate completions for the given prefix.

        This takes the most-recently-received set of active nameplates from
        the rendezvous server, finds the subset that start with the given
        prefix, and returns the result.  The result strings include the
        prefix and the terminating hyphen, in random order.

        This returns synchronously: it does not wait for a server response.
        If called before getting any response from the server, it will return
        an empty set.  If user input causes completion, it may be a good idea
        to kick off a new ``refresh_nameplates()`` too, in case the user is
        bouncing on the TAB key in the hopes of seeing their expected
        nameplate appear in the list eventually.

        :param str prefix: the nameplate as typed so far

        :return: a set of potential completions
        :rtype: set[str]
        """

    def choose_nameplate(nameplate):
        """
        Commit to a nameplate, allowing the word-completion phase to begin.

        This may only be called once.  Calling it a second time will raise
        ``AlreadyChoseNameplateError``.

        :param str nameplate: the complete nameplate, without a trailing
            hyphen

        :return: None
        """

    def when_wordlist_is_available():
        """
        Wait for the wordlist to be available.

        This fires when the wordlist is available, which means
        ``get_word_completions()`` is able to return a non-empty set.  This
        requires the nameplate be submitted, and may also require some server
        interaction (to claim the channel and learn a channel-specific
        wordlist, e.g. for i18n language selection).

        :return: a ``Deferred`` that fires when the wordlist is available
        :rtype: Deferred[None]
        """

    def get_word_completions(prefix):
        """
        Return a list of word completions for the given prefix.

        This takes the claimed channel's wordlist, finds the subset that
        start with the given prefix, and returns the result.  The result
        strings include the prefix and the terminating hyphen, in random
        order.

        The prefix should not include the nameplate, but should include
        whatever words have been selected so far (the default uses separate
        odd/even wordlists, which means the completion for a single string
        depends upon how many words have been entered so far).

        This returns synchronously: it does not wait for a server response.
        If called before getting the wordlist, it will return an empty set.

        If called before ``choose_nameplate()``, this will raise
        ``MustChooseNameplateFirstError``.  If called after
        ``choose_words()``, this will raise ``AlreadyChoseWordsError``.

        :param str prefix: the words typed so far

        :return: a set of potential completions
        :rtype: set[str]
        """

    def choose_words(words):
        """
        Submit the final words.

        This should be called when the user is finished typing in the code,
        and terminates the code-entry process.  It does not return anything,
        but will cause the Wormhole's ``w.get_code()`` to fire, and initiates
        the wormhole connection process.

        It accepts a string like "purple-sausages", without the leading
        nameplate (which must have been submitted to ``choose_nameplate()``
        earlier) or its hyphen.  If ``choose_nameplate()`` was not called
        first, this will raise ``MustChooseNameplateFirstError``.

        This may only be called once, otherwise ``AlreadyChoseWordsError``
        will be raised.

        :param str words: the 'words' portion of the wormhole code

        :return: None
        """


class IJournal(Interface):  # TODO: this needs to be public
    pass


class IDilator(Interface):
    pass


class IDilationManager(Interface):
    pass


class IDilationConnector(Interface):
    pass


class ISubChannel(Interface):
    pass


class IInbound(Interface):
    pass


class IOutbound(Interface):
    pass



import sys

from attr import attrib, attrs
from twisted.internet.defer import inlineCallbacks
from twisted.internet.endpoints import clientFromString
from zope.interface.declarations import directlyProvides

from . import _interfaces, errors
from .timing import DebugTiming

try:
    import txtorcon
except ImportError:
    txtorcon = None


@attrs
class SocksOnlyTor:
    _reactor = attrib()

    def stream_via(self, host, port, tls=False):
        return txtorcon.TorClientEndpoint(
            host,
            port,
            socks_endpoint=None,  # tries localhost:9050 and 9150
            tls=tls,
            reactor=self._reactor,
        )


@inlineCallbacks
def get_tor(reactor,
            launch_tor=False,
            tor_control_port=None,
            timing=None,
            stderr=sys.stderr):
    """
    If launch_tor=True, I will try to launch a new Tor process, ask it
    for its SOCKS and control ports, and use those for outbound
    connections (and inbound onion-service listeners, if necessary).

    Otherwise if tor_control_port is provided, I will attempt to connect
    to an existing Tor's control port at the endpoint it specifies.  I'll
    ask that Tor for its SOCKS port.

    With no arguments, I will try to connect to an existing Tor's control
    port at the usual places: [unix:/var/run/tor/control,
    tcp:127.0.0.1:9051, tcp:127.0.0.1:9151].  If any are successful, I'll
    ask that Tor for its SOCKS port.  If none are successful, I'll
    attempt to do SOCKS to the usual places: [tcp:127.0.0.1:9050,
    tcp:127.0.0.1:9150].

    If I am unable to make a SOCKS connection, the initial connection to
    the Rendezvous Server will fail, and the program will terminate.

    Control-port connections can only succeed if I can authenticate (by
    reading a cookie file named by the Tor process), so the current user
    must have permission to read that file (either they started Tor, e.g.
    TorBrowser, or they are in a unix group that's been given access,
    e.g. debian-tor).
    """
    # rationale: launching a new Tor takes a long time, so only do it if
    # the user specifically asks for it with --launch-tor. Using an
    # existing Tor should be much faster, but still requires general
    # permission via --tor.

    if not txtorcon:
        raise errors.NoTorError()

    if not isinstance(launch_tor, bool):  # note: False is int
        raise TypeError("launch_tor= must be boolean")
    if not isinstance(tor_control_port, (str, type(None))):
        raise TypeError("tor_control_port= must be str or None")
    assert tor_control_port != ""
    if launch_tor and tor_control_port is not None:
        raise ValueError("cannot combine --launch-tor and --tor-control-port=")
    timing = timing or DebugTiming()

    # Connect to an existing Tor, or create a new one. If we need to
    # launch an onion service, then we need a working control port (and
    # authentication cookie). If we're only acting as a client, we don't
    # need the control port.

    if launch_tor:
        print(
            " launching a new Tor process, this may take a while..",
            file=stderr)
        with timing.add("launch tor"):
            tor = yield txtorcon.launch(reactor,
                                        # data_directory=,
                                        # tor_binary=,
                                        )
    elif tor_control_port:
        with timing.add("find tor"):
            control_ep = clientFromString(reactor, tor_control_port)
            tor = yield txtorcon.connect(reactor, control_ep)  # might raise
            print(
                f" using Tor via control port at {tor_control_port}",
                file=stderr)
    else:
        # Let txtorcon look through a list of usual places. If that fails,
        # we'll arrange to attempt the default SOCKS port
        with timing.add("find tor"):
            try:
                tor = yield txtorcon.connect(reactor)
                print(" using Tor via default control port", file=stderr)
            except Exception:
                # TODO: make this more specific. I think connect() is
                # likely to throw a reactor.connectTCP -type error, like
                # ConnectionFailed or ConnectionRefused or something
                print(
                    " unable to find default Tor control port, using SOCKS",
                    file=stderr)
                tor = SocksOnlyTor(reactor)
    directlyProvides(tor, _interfaces.ITorManager)
    return tor



import os
from urllib.parse import urlparse
from attr import attrs, attrib
from attr.validators import instance_of, optional
from zope.interface import implementer
from twisted.python import log
from twisted.internet import defer, endpoints, task
from twisted.application import internet
from autobahn.twisted import websocket
from . import _interfaces, errors
from .util import (bytes_to_hexstr, hexstr_to_bytes, bytes_to_dict,
                   dict_to_bytes, provides)

from ._status import Connecting, Connected


class WSClient(websocket.WebSocketClientProtocol):
    def onConnect(self, response):
        # this fires during WebSocket negotiation, and isn't very useful
        # unless you want to modify the protocol settings
        pass

    def onOpen(self, *args):
        # this fires when the WebSocket is ready to go. No arguments
        # self.wormhole_open = True
        self._RC.ws_open(self)

    def onMessage(self, payload, isBinary):
        assert not isBinary
        try:
            self._RC.ws_message(payload)
        except Exception:
            from twisted.python.failure import Failure
            print("LOGGING", Failure())
            log.err()
            raise

    def onClose(self, wasClean, code, reason):
        self._RC.ws_close(wasClean, code, reason)
        # if self.wormhole_open:
        #     self.wormhole._ws_closed(wasClean, code, reason)
        # else:
        #     # we closed before establishing a connection (onConnect) or
        #     # finishing WebSocket negotiation (onOpen): errback
        #     self.factory.d.errback(error.ConnectError(reason))


class WSFactory(websocket.WebSocketClientFactory):
    protocol = WSClient

    def __init__(self, RC, *args, **kwargs):
        websocket.WebSocketClientFactory.__init__(self, *args, **kwargs)
        self._RC = RC

    def buildProtocol(self, addr):
        proto = websocket.WebSocketClientFactory.buildProtocol(self, addr)
        proto._RC = self._RC
        # proto.wormhole_open = False
        return proto


@attrs
@implementer(_interfaces.IRendezvousConnector)
class RendezvousConnector:
    _url = attrib(validator=instance_of(str))
    _appid = attrib(validator=instance_of(str))
    _side = attrib(validator=instance_of(str))
    _reactor = attrib()
    _journal = attrib(validator=provides(_interfaces.IJournal))
    _tor = attrib(validator=optional(provides(_interfaces.ITorManager)))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    _client_version = attrib(validator=instance_of(tuple))
    _evolve_status = attrib()  # callable taking kwargs to change aspects of the status

    def __attrs_post_init__(self):
        self._have_made_a_successful_connection = False
        self._stopping = False

        self._trace = None
        self._ws = None
        f = WSFactory(self, self._url)
        # kind-of match what Dilation does for peer connections;
        # there, we send a ping every 30s and give up on the
        # connection if two fail in a row -- autobahn doesn't give us
        # _quite_ those same hooks, but we time out in 60s which will
        # be similar behavior.
        f.setProtocolOptions(autoPingInterval=30, autoPingTimeout=60)
        ep = self._make_endpoint(self._url)

        # ideally, Twisted's ClientService would have an API to tell
        # us when it tries to do a connection, but it doesn't. So
        # instead, we wrap the endpoint's "connect()" method so we
        # know a connection attempt has been made
        orig = ep.connect
        def connect_wrap(*args, **kw):
            self._evolve_status(mailbox_connection=Connecting(self._url, self._reactor.seconds()))
            return orig(*args, **kw)
        ep.connect = connect_wrap

        self._connector = internet.ClientService(ep, f)
        faf = None if self._have_made_a_successful_connection else 1
        d = self._connector.whenConnected(failAfterFailures=faf)
        # if the initial connection fails, signal an error and shut down. do
        # this in a different reactor turn to avoid some hazards

        d.addBoth(lambda res: task.deferLater(self._reactor, 0.0, lambda: res))
        # TODO: use EventualQueue
        d.addErrback(self._initial_connection_failed)
        self._debug_record_inbound_f = None

    def set_trace(self, f):
        self._trace = f

    def _debug(self, what):
        if self._trace:
            self._trace(old_state="", input=what, new_state="")

    def _make_endpoint(self, url):
        p = urlparse(url)
        tls = (p.scheme == "wss")
        port = p.port or (443 if tls else 80)
        if self._tor:
            return self._tor.stream_via(p.hostname, port, tls=tls)
        if tls:
            return endpoints.clientFromString(self._reactor,
                                              f"tls:{p.hostname}:{port}")
        return endpoints.HostnameEndpoint(self._reactor, p.hostname, port)

    def wire(self, boss, nameplate, mailbox, allocator, lister, terminator):
        self._B = _interfaces.IBoss(boss)
        self._N = _interfaces.INameplate(nameplate)
        self._M = _interfaces.IMailbox(mailbox)
        self._A = _interfaces.IAllocator(allocator)
        self._L = _interfaces.ILister(lister)
        self._T = _interfaces.ITerminator(terminator)

    # from Boss
    def start(self):
        self._connector.startService()

    # from Mailbox
    def tx_claim(self, nameplate):
        self._tx("claim", nameplate=nameplate)

    def tx_open(self, mailbox):
        self._tx("open", mailbox=mailbox)

    def tx_add(self, phase, body):
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        self._tx("add", phase=phase, body=bytes_to_hexstr(body))

    def tx_release(self, nameplate):
        self._tx("release", nameplate=nameplate)

    def tx_close(self, mailbox, mood):
        self._tx("close", mailbox=mailbox, mood=mood)

    def stop(self):
        # ClientService.stopService is defined to "Stop attempting to
        # reconnect and close any existing connections"
        self._stopping = True  # to catch _initial_connection_failed error
        d = defer.maybeDeferred(self._connector.stopService)
        # ClientService.stopService always fires with None, even if the
        # initial connection failed, so log.err just in case
        d.addErrback(log.err)
        d.addBoth(self._stopped)

    # from Lister
    def tx_list(self):
        self._tx("list")

    # from Code
    def tx_allocate(self):
        self._tx("allocate")

    # from our ClientService
    def _initial_connection_failed(self, f):
        if not self._stopping:
            sce = errors.ServerConnectionError(self._url, f.value)
            d = defer.maybeDeferred(self._connector.stopService)
            # this should happen right away: the ClientService ought to be in
            # the "_waiting" state, and everything in the _waiting.stop
            # transition is immediate
            d.addErrback(log.err)  # just in case something goes wrong
            d.addCallback(lambda _: self._B.error(sce))

    # from our WSClient (the WebSocket protocol)
    def ws_open(self, proto):
        self._debug("R.connected")
        self._have_made_a_successful_connection = True
        self._evolve_status(mailbox_connection=Connected(self._url))
        self._ws = proto
        try:
            self._tx(
                "bind",
                appid=self._appid,
                side=self._side,
                client_version=self._client_version)
            self._N.connected()
            self._M.connected()
            self._L.connected()
            self._A.connected()
        except Exception as e:
            self._B.error(e)
            raise
        self._debug("R.connected finished notifications")

    def ws_message(self, payload):
        msg = bytes_to_dict(payload)
        if msg["type"] != "ack":
            self._debug("R.rx({} {}{})".format(
                msg["type"],
                msg.get("phase", ""),
                "[mine]" if msg.get("side", "") == self._side else "",
            ))

        self._timing.add("ws_receive", _side=self._side, message=msg)
        if self._debug_record_inbound_f:
            self._debug_record_inbound_f(msg)
        mtype = msg["type"]
        meth = getattr(self, "_response_handle_" + mtype, None)
        if not meth:
            # make tests fail, but real application will ignore it
            log.err(
                errors._UnknownMessageTypeError(
                    f"Unknown inbound message type {msg!r}"))
            return
        try:
            return meth(msg)
        except Exception as e:
            log.err(e)
            self._B.error(e)
            raise

    def ws_close(self, wasClean, code, reason):
        self._debug("R.lost")
        was_open = bool(self._ws)
        self._ws = None
        # when Autobahn connects to a non-websocket server, it gets a
        # CLOSE_STATUS_CODE_ABNORMAL_CLOSE, and delivers onClose() without
        # ever calling onOpen first. This confuses our state machines, so
        # avoid telling them we've lost the connection unless we'd previously
        # told them we'd connected.
        if was_open:
            self._N.lost()
            self._M.lost()
            self._L.lost()
            self._A.lost()

        # and if this happens on the very first connection, then we treat it
        # as a failed initial connection, even though ClientService didn't
        # notice it. There's a Twisted ticket (#8375) about giving
        # ClientService an extra setup function to use, so it can tell
        # whether post-connection negotiation was successful or not, and
        # restart the process if it fails. That would be useful here, so that
        # failAfterFailures=1 would do the right thing if the initial TCP
        # connection succeeds but the first WebSocket negotiation fails.
        if not self._have_made_a_successful_connection:
            # shut down the ClientService, which currently thinks it has a
            # valid connection
            sce = errors.ServerConnectionError(self._url, reason)
            d = defer.maybeDeferred(self._connector.stopService)
            d.addErrback(log.err)  # just in case something goes wrong
            # tell the Boss to quit and inform the user
            d.addCallback(lambda _: self._B.error(sce))

    # internal
    def _stopped(self, res):
        self._T.stoppedRC()

    def _tx(self, mtype, **kwargs):
        assert self._ws
        # msgid is used by misc/dump-timing.py to correlate our sends with
        # their receives, and vice versa. They are also correlated with the
        # ACKs we get back from the server (which we otherwise ignore). There
        # are so few messages, 16 bits is enough to be mostly-unique.
        kwargs["id"] = bytes_to_hexstr(os.urandom(2))
        kwargs["type"] = mtype
        self._debug(f"R.tx({mtype.upper()} {kwargs.get('phase', '')})")
        payload = dict_to_bytes(kwargs)
        self._timing.add("ws_send", _side=self._side, **kwargs)
        self._ws.sendMessage(payload, False)
        # might be nice to have a "debug" hook here to track all
        # messages sent to the mailbox, with timestamps

    def _response_handle_allocated(self, msg):
        nameplate = msg["nameplate"]
        assert isinstance(nameplate, str), type(nameplate)
        self._A.rx_allocated(nameplate)

    def _response_handle_nameplates(self, msg):
        # we get list of {id: ID}, with maybe more attributes in the future
        nameplates = msg["nameplates"]
        assert isinstance(nameplates, list), type(nameplates)
        nids = set()
        for n in nameplates:
            assert isinstance(n, dict), type(n)
            nameplate_id = n["id"]
            assert isinstance(nameplate_id, str), type(nameplate_id)
            nids.add(nameplate_id)
        # deliver a set of nameplate ids
        self._L.rx_nameplates(nids)

    def _response_handle_ack(self, msg):
        pass

    def _response_handle_error(self, msg):
        # the server sent us a type=error. Most cases are due to our mistakes
        # (malformed protocol messages, sending things in the wrong order),
        # but it can also result from CrowdedError (more than two clients
        # using the same channel).
        err = msg["error"]
        orig = msg["orig"]
        self._B.rx_error(err, orig)

    def _response_handle_welcome(self, msg):
        self._B.rx_welcome(msg["welcome"])

    def _response_handle_claimed(self, msg):
        mailbox = msg["mailbox"]
        assert isinstance(mailbox, str), type(mailbox)
        self._N.rx_claimed(mailbox)

    def _response_handle_message(self, msg):
        side = msg["side"]
        phase = msg["phase"]
        assert isinstance(phase, str), type(phase)
        body = hexstr_to_bytes(msg["body"])  # bytes
        self._M.rx_message(side, phase, body)

    def _response_handle_released(self, msg):
        self._N.rx_released()

    def _response_handle_closed(self, msg):
        self._M.rx_closed()

    # record, message, payload, packet, bundle, ciphertext, plaintext



import unicodedata

from unittest import mock

from .. import util

def test_to_bytes():
    b = util.to_bytes("abc")
    assert isinstance(b, bytes)
    assert b == b"abc"

    A = unicodedata.lookup("LATIN SMALL LETTER A WITH DIAERESIS")
    b = util.to_bytes(A + "bc")
    assert isinstance(b, bytes)
    assert b == b"\xc3\xa4\x62\x63"

def test_bytes_to_hexstr():
    b = b"\x00\x45\x91\xfe\xff"
    hexstr = util.bytes_to_hexstr(b)
    assert isinstance(hexstr, str)
    assert hexstr == "004591feff"

def test_hexstr_to_bytes():
    hexstr = "004591feff"
    b = util.hexstr_to_bytes(hexstr)
    hexstr = util.bytes_to_hexstr(b)
    assert isinstance(b, bytes)
    assert b == b"\x00\x45\x91\xfe\xff"

def test_dict_to_bytes():
    d = {"a": "b"}
    b = util.dict_to_bytes(d)
    assert isinstance(b, bytes)
    assert b == b'{"a": "b"}'

def test_bytes_to_dict():
    b = b'{"a": "b", "c": 2}'
    d = util.bytes_to_dict(b)
    assert isinstance(d, dict)
    assert d == {"a": "b", "c": 2}


def test_free_space():
    free = util.estimate_free_space(".")
    assert isinstance(free, (int, type(None))), repr(free)
    # some platforms (I think the VMs used by travis are in this
    # category) return 0, and windows will return None, so don't assert
    # anything more specific about the return value

def test_no_statvfs():
    # this mock.patch fails on windows, which is sad because windows is
    # the one platform that the code under test was supposed to help with
    try:
        with mock.patch("os.statvfs", side_effect=AttributeError()):
            assert util.estimate_free_space(".") is None
    except AttributeError:  # raised by mock.get_original()
        pass



import io
import os

from unittest import mock

from ..cli import cmd_ssh
import pytest

OTHERS = ["config", "config~", "known_hosts", "known_hosts~"]


def test_find_one():
    files = OTHERS + ["id_rsa.pub", "id_rsa"]
    pubkey_data = "ssh-rsa AAAAkeystuff email@host\n"
    pubkey_file = io.StringIO(pubkey_data)
    with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
        with mock.patch("os.listdir", return_value=files) as ld:
            with mock.patch(
                    "wormhole.cli.cmd_ssh.open", return_value=pubkey_file):
                res = cmd_ssh.find_public_key()
    assert ld.mock_calls == \
                     [mock.call(os.path.expanduser("~/.ssh/"))]
    assert len(res) == 3, res
    kind, keyid, pubkey = res
    assert kind == "ssh-rsa"
    assert keyid == "email@host"
    assert pubkey == pubkey_data

def test_find_none():
    files = OTHERS  # no pubkey
    with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
        with mock.patch("os.listdir", return_value=files):
            with pytest.raises(cmd_ssh.PubkeyError) as f:
                cmd_ssh.find_public_key()
    dot_ssh = os.path.expanduser("~/.ssh/")
    assert str(f.value) == f"No public keys in '{dot_ssh}'"

def test_bad_hint():
    with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=False):
        with pytest.raises(cmd_ssh.PubkeyError) as f:
            cmd_ssh.find_public_key(hint="bogus/path")
    assert str(f.value) == "Can't find 'bogus/path'"

def test_find_multiple():
    files = OTHERS + ["id_rsa.pub", "id_rsa", "id_dsa.pub", "id_dsa"]
    pubkey_data = "ssh-rsa AAAAkeystuff email@host\n"
    pubkey_file = io.StringIO(pubkey_data)
    with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
        with mock.patch("os.listdir", return_value=files):
            responses = iter(["frog", "NaN", "-1", "0"])
            with mock.patch(
                    "click.prompt", side_effect=lambda p: next(responses)):
                with mock.patch(
                        "wormhole.cli.cmd_ssh.open",
                        return_value=pubkey_file):
                    res = cmd_ssh.find_public_key()
    assert len(res) == 3, res
    kind, keyid, pubkey = res
    assert kind == "ssh-rsa"
    assert keyid == "email@host"
    assert pubkey == pubkey_data

def test_comment_with_spaces():
    files = OTHERS + ["id_ed25519.pub", "id_ed25519"]
    pubkey_data = "ssh-ed25519 AAAAkeystuff comment with spaces"
    pubkey_file = io.StringIO(pubkey_data)

    with mock.patch("wormhole.cli.cmd_ssh.exists", return_value=True):
        with mock.patch("os.listdir", return_value=files) as ld:
            with mock.patch(
                    "wormhole.cli.cmd_ssh.open", return_value=pubkey_file):
                res = cmd_ssh.find_public_key()
    assert ld.mock_calls == \
                     [mock.call(os.path.expanduser("~/.ssh/"))]
    assert len(res) == 3, res
    kind, keyid, pubkey = res
    assert kind == "ssh-ed25519"
    assert keyid == "comment with spaces"
    assert pubkey == pubkey_data



from attr import attrib, attrs
from attr.validators import instance_of
from automat import MethodicalMachine
from zope.interface import implementer

from . import _interfaces
from ._key import CryptoError, decrypt_data, derive_key, derive_phase_key
from .util import provides


@attrs
@implementer(_interfaces.IReceive)
class Receive:
    _side = attrib(validator=instance_of(str))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace",
                        lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._key = None

    def wire(self, boss, send):
        self._B = _interfaces.IBoss(boss)
        self._S = _interfaces.ISend(send)

    @m.state(initial=True)
    def S0_unknown_key(self):
        pass  # pragma: no cover

    @m.state()
    def S1_unverified_key(self):
        pass  # pragma: no cover

    @m.state()
    def S2_verified_key(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def S3_scared(self):
        pass  # pragma: no cover

    # from Ordering
    def got_message(self, side, phase, body):
        assert isinstance(side, str), type(phase)
        assert isinstance(phase, str), type(phase)
        assert isinstance(body, bytes), type(body)
        assert self._key
        data_key = derive_phase_key(self._key, side, phase)
        try:
            plaintext = decrypt_data(data_key, body)
        except CryptoError:
            self.got_message_bad()
            return
        self.got_message_good(phase, plaintext)

    @m.input()
    def got_message_good(self, phase, plaintext):
        pass

    @m.input()
    def got_message_bad(self):
        pass

    # from Key
    @m.input()
    def got_key(self, key):
        pass

    @m.output()
    def record_key(self, key):
        self._key = key

    @m.output()
    def S_got_verified_key(self, phase, plaintext):
        assert self._key
        self._S.got_verified_key(self._key)

    @m.output()
    def W_happy(self, phase, plaintext):
        self._B.happy()

    @m.output()
    def W_got_verifier(self, phase, plaintext):
        self._B.got_verifier(derive_key(self._key, b"wormhole:verifier"))

    @m.output()
    def W_got_message(self, phase, plaintext):
        assert isinstance(phase, str), type(phase)
        assert isinstance(plaintext, bytes), type(plaintext)
        self._B.got_message(phase, plaintext)

    @m.output()
    def W_scared(self):
        self._B.scared()

    S0_unknown_key.upon(got_key, enter=S1_unverified_key, outputs=[record_key])
    S1_unverified_key.upon(
        got_message_good,
        enter=S2_verified_key,
        outputs=[S_got_verified_key, W_happy, W_got_verifier, W_got_message])
    S1_unverified_key.upon(
        got_message_bad, enter=S3_scared, outputs=[W_scared])
    S2_verified_key.upon(got_message_bad, enter=S3_scared, outputs=[W_scared])
    S2_verified_key.upon(
        got_message_good, enter=S2_verified_key, outputs=[W_got_message])
    S3_scared.upon(got_message_good, enter=S3_scared, outputs=[])
    S3_scared.upon(got_message_bad, enter=S3_scared, outputs=[])



from unittest import mock
from zope.interface import directlyProvides, implementer
from twisted.internet.interfaces import ITransport, IHalfCloseableProtocol, IProtocol
from twisted.internet.error import ConnectionDone
from ..._interfaces import IDilationManager
from ..._dilation.subchannel import (SubChannel,
                                     _WormholeAddress, SubchannelAddress,
                                     AlreadyClosedError,
                                     NormalCloseUsedOnHalfCloseable,
                                     SubchannelDemultiplex,
                                     UnexpectedSubprotocol)
from ..._dilation.manager import Once
from .common import mock_manager
import pytest
from hypothesis import given
from hypothesis.strategies import text, lists


def make_sc(set_protocol=True, half_closeable=False):
    scid = 4
    hostaddr = _WormholeAddress()
    peeraddr = SubchannelAddress("proto")
    m = mock_manager()
    sc = SubChannel(scid, m, hostaddr, peeraddr)
    p = mock.Mock()
    if half_closeable:
        directlyProvides(p, IHalfCloseableProtocol)
    if set_protocol:
        sc._set_protocol(p)
    return sc, m, scid, hostaddr, peeraddr, p


def test_subchannel_once():
    o = Once(ValueError)
    o()
    with pytest.raises(ValueError):
        o()


def test_subchannel_create():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()
    assert ITransport.providedBy(sc)
    assert m.mock_calls == []
    assert sc.getHost() is hostaddr
    assert sc.getPeer() is peeraddr


def test_subchannel_write():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    sc.write(b"data")
    assert m.mock_calls == [mock.call.send_data(scid, b"data")]
    m.mock_calls[:] = []
    sc.writeSequence([b"more", b"data"])
    assert m.mock_calls == [mock.call.send_data(scid, b"moredata")]


def test_subchannel_write_when_closing():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    sc.loseConnection()
    assert m.mock_calls == [mock.call.send_close(scid)]
    m.mock_calls[:] = []

    with pytest.raises(AlreadyClosedError) as e:
        sc.write(b"data")
    assert str(e.value) == "write not allowed on closed subchannel"


def test_subchannel_local_close():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    sc.loseConnection()
    assert m.mock_calls == [mock.call.send_close(scid)]
    m.mock_calls[:] = []

    # late arriving data is still delivered
    sc.remote_data(b"late")
    assert p.mock_calls == [mock.call.dataReceived(b"late")]
    p.mock_calls[:] = []

    sc.remote_close()
    assert_connectionDone(p.mock_calls)


def test_subchannel_local_double_close():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    sc.loseConnection()
    assert m.mock_calls == [mock.call.send_close(scid)]
    m.mock_calls[:] = []

    with pytest.raises(AlreadyClosedError) as e:
        sc.loseConnection()
    assert str(e.value) == "loseConnection not allowed on closed subchannel"

def assert_connectionDone(mock_calls):
    assert len(mock_calls) == 1
    assert mock_calls[0][0] == "connectionLost"
    assert len(mock_calls[0][1]) == 1
    assert isinstance(mock_calls[0][1][0], ConnectionDone)


def test_subchannel_remote_close():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()
    sc.remote_close()
    assert m.mock_calls == [mock.call.send_close(scid),
                                    mock.call.subchannel_closed(scid, sc)]
    assert_connectionDone(p.mock_calls)


def test_subchannel_data():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()
    sc.remote_data(b"data")
    assert p.mock_calls == [mock.call.dataReceived(b"data")]
    p.mock_calls[:] = []
    sc.remote_data(b"not")
    sc.remote_data(b"coalesced")
    assert p.mock_calls == [mock.call.dataReceived(b"not"),
                                    mock.call.dataReceived(b"coalesced"),
                                    ]


def test_subchannel_data_before_open():
    sc, m, scid, hostaddr, peeraddr, p = make_sc(set_protocol=False)
    sc.remote_data(b"data1")
    sc.remote_data(b"data2")
    assert p.mock_calls == []
    sc._set_protocol(p)
    sc._deliver_queued_data()
    assert p.mock_calls == [mock.call.dataReceived(b"data1"),
                                    mock.call.dataReceived(b"data2")]
    p.mock_calls[:] = []
    sc.remote_data(b"more")
    assert p.mock_calls == [mock.call.dataReceived(b"more")]


def test_subchannel_close_before_open():
    sc, m, scid, hostaddr, peeraddr, p = make_sc(set_protocol=False)
    sc.remote_close()
    assert p.mock_calls == []
    sc._set_protocol(p)
    sc._deliver_queued_data()
    assert_connectionDone(p.mock_calls)


def test_subchannel_producer():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    sc.pauseProducing()
    assert m.mock_calls == [mock.call.subchannel_pauseProducing(sc)]
    m.mock_calls[:] = []
    sc.resumeProducing()
    assert m.mock_calls == [mock.call.subchannel_resumeProducing(sc)]
    m.mock_calls[:] = []
    sc.stopProducing()
    assert m.mock_calls == [mock.call.subchannel_stopProducing(sc)]
    m.mock_calls[:] = []


def test_subchannel_consumer():
    sc, m, scid, hostaddr, peeraddr, p = make_sc()

    # TODO: more, once this is implemented
    sc.registerProducer(None, True)
    sc.unregisterProducer()


def test_halfcloseable_create():
    sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)
    assert ITransport.providedBy(sc)
    assert m.mock_calls == []
    assert sc.getHost() is hostaddr
    assert sc.getPeer() is peeraddr


def test_halfcloseable_local_close():
    sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)

    sc.write(b"data")
    assert m.mock_calls == [mock.call.send_data(scid, b"data")]
    m.mock_calls[:] = []
    sc.writeSequence([b"more", b"data"])
    assert m.mock_calls == [mock.call.send_data(scid, b"moredata")]
    m.mock_calls[:] = []

    sc.remote_data(b"inbound1")
    assert p.mock_calls == [mock.call.dataReceived(b"inbound1")]
    p.mock_calls[:] = []

    with pytest.raises(NormalCloseUsedOnHalfCloseable) as e:
        sc.loseConnection()  # TODO: maybe this shouldn't be an error

    # after a local close, we can't write anymore, but we can still
    # receive data
    sc.loseWriteConnection()  # TODO or loseConnection?
    assert m.mock_calls == [mock.call.send_close(scid)]
    m.mock_calls[:] = []
    assert p.mock_calls == [mock.call.writeConnectionLost()]
    p.mock_calls[:] = []

    with pytest.raises(AlreadyClosedError) as e:
        sc.write(b"data")
    assert str(e.value) == "write not allowed on closed subchannel"

    with pytest.raises(AlreadyClosedError) as e:
        sc.loseWriteConnection()
    assert str(e.value) == "loseConnection not allowed on closed subchannel"

    with pytest.raises(NormalCloseUsedOnHalfCloseable) as e:
        sc.loseConnection()  # TODO: maybe expect AlreadyClosedError

    sc.remote_data(b"inbound2")
    assert p.mock_calls == [mock.call.dataReceived(b"inbound2")]
    p.mock_calls[:] = []

    # the remote end will finally shut down the connection
    sc.remote_close()
    assert m.mock_calls == [mock.call.subchannel_closed(scid, sc)]
    assert p.mock_calls == [mock.call.readConnectionLost()]


def test_halfcloseable_remote_close():
    sc, m, scid, hostaddr, peeraddr, p = make_sc(half_closeable=True)

    sc.write(b"data")
    assert m.mock_calls == [mock.call.send_data(scid, b"data")]
    m.mock_calls[:] = []

    sc.remote_data(b"inbound1")
    assert p.mock_calls == [mock.call.dataReceived(b"inbound1")]
    p.mock_calls[:] = []

    # after a remote close, we can still write data
    sc.remote_close()
    assert m.mock_calls == []
    assert p.mock_calls == [mock.call.readConnectionLost()]
    p.mock_calls[:] = []

    sc.write(b"out2")
    assert m.mock_calls == [mock.call.send_data(scid, b"out2")]
    m.mock_calls[:] = []

    # and a local close will shutdown the connection
    sc.loseWriteConnection()
    assert m.mock_calls == [mock.call.send_close(scid),
                                    mock.call.subchannel_closed(scid, sc)]
    assert p.mock_calls == [mock.call.writeConnectionLost()]


# these fakes (used by the below tests) should be expanded for use
# elsewhere if required -- or perhaps we can make the "real" objects
# easier to construct for tests.


@implementer(IDilationManager)
class FakeManager:
    pass


@implementer(IProtocol)
class FakeProtocol:
    transport = None
    def makeConnection(self, transport):
        self.transport = transport


class FakeFactory:
    def __init__(self):
        self.builds = []

    def buildProtocol(self, addr):
        self.builds.append(addr)
        return FakeProtocol()


def test_demultiplex():
    """
    Confirm we hold multiple open calls until a listener appears
    """
    demult = SubchannelDemultiplex()

    fake_manager = FakeManager()
    hostaddr = _WormholeAddress()

    addr = SubchannelAddress("subproto")
    t0 = SubChannel(0, fake_manager, hostaddr, addr)
    t1 = SubChannel(1, fake_manager, hostaddr, addr)

    factory = FakeFactory()

    demult._got_open(t0, addr)
    demult._got_open(t1, addr)

    # nothing listening yet, so we should have accumulated two opens
    demult.register("subproto", factory)

    # now we should have gotten two protocol builds (i.e. after we
    # "listen" with our factory)
    assert factory.builds == [addr, addr]


@given(
    lists(
        text(min_size=1),
        min_size=1
    )
)
def test_demultiplex_multiple_protocols(protocols):
    """
    Interleave two separate subprotocols and confirm we hold them open
    until a following listen
    """
    # Hypothesis is giving us a list of subchannel names to open
    # protocols with. It would be best if we had lots of repeats, but
    # trying to _force_ that with assume() makes Hypothesis grumpy
    demult = SubchannelDemultiplex()
    fake_manager = FakeManager()
    hostaddr = _WormholeAddress()
    unique_protocols = set(protocols)

    addresses = {
        nm: SubchannelAddress(nm)
        for nm in unique_protocols
    }
    subchannels = [
        SubChannel(i, fake_manager, hostaddr, addresses[nm])
        for i, nm in enumerate(protocols)
    ]

    for nm, sub in zip(protocols, subchannels):
        demult._got_open(sub, addresses[nm])

    # once we listen, each kind of subprotocol should have represented
    # twice
    factories = {
        nm: FakeFactory()
        for nm in unique_protocols
    }
    for nm, factory in factories.items():
        demult.register(nm, factory)

    # each factory should have had buildProtocol invoked once for each
    # incoming open we had for that name
    for nm, factory in factories.items():
        assert factory.builds == [addresses[nm]] * protocols.count(nm)


def test_unexpected_protocol():
    """
    An incoming subprotocol produces an error if we specify a set of
    expected protocols (and the incoming one is not on the list)
    """
    demult = SubchannelDemultiplex(expected_subprotocols=["foo"])
    fake_manager = FakeManager()
    hostaddr = _WormholeAddress()

    addr = SubchannelAddress("subproto")
    t0 = SubChannel(0, fake_manager, hostaddr, addr)
    with pytest.raises(UnexpectedSubprotocol):
        demult._got_open(t0, addr)



from ..._dilation.encode import to_be4, from_be4
import pytest


def test_be4():
    assert to_be4(0) == b"\x00\x00\x00\x00"
    assert to_be4(1) == b"\x00\x00\x00\x01"
    assert to_be4(256) == b"\x00\x00\x01\x00"
    assert to_be4(257) == b"\x00\x00\x01\x01"
    with pytest.raises(ValueError):
        to_be4(-1)
    with pytest.raises(ValueError):
        to_be4(2**32)

    assert from_be4(b"\x00\x00\x00\x00") == 0
    assert from_be4(b"\x00\x00\x00\x01") == 1
    assert from_be4(b"\x00\x00\x01\x00") == 256
    assert from_be4(b"\x00\x00\x01\x01") == 257

    with pytest.raises(TypeError):
        from_be4(0)
    with pytest.raises(ValueError):
        from_be4(b"\x01\x00\x00\x00\x00")



from collections import deque
from attr import attrs, attrib
from zope.interface import implementer
from twisted.internet.interfaces import IPushProducer, IPullProducer
from twisted.python import log
from twisted.python.reflect import safe_str
from .._interfaces import IDilationManager, IOutbound
from ..util import provides
from .connection import KCM, Ping, Pong, Ack


# Outbound flow control: app writes to subchannel, we write to Connection

# The app can register an IProducer of their choice, to let us throttle their
# outbound data. Not all subchannels will have producers registered, and the
# producer probably won't be the IProtocol instance (it'll be something else
# which feeds data out through the protocol, like a t.p.basic.FileSender). If
# a producerless subchannel writes too much, we won't be able to stop them,
# and we'll keep writing records into the Connection even though it's asked
# us to pause. Likewise, when the connection is down (and we're busily trying
# to reestablish a new one), registered subchannels will be paused, but
# unregistered ones will just dump everything in _outbound_queue, and we'll
# consume memory without bound until they stop.

# We need several things:
#
# * Add each registered IProducer to a list, whose order remains stable. We
#   want fairness under outbound throttling: each time the outbound
#   connection opens up (our resumeProducing method is called), we should let
#   just one producer have an opportunity to do transport.write, and then we
#   should pause them again, and not come back to them until everyone else
#   has gotten a turn. So we want an ordered list of producers to track this
#   rotation.
#
# * Remove the IProducer if/when the protocol uses unregisterProducer
#
# * Remove any registered IProducer when the associated Subchannel is closed.
#   This isn't a problem for normal transports, because usually there's a
#   one-to-one mapping from Protocol to Transport, so when the Transport you
#   forget the only reference to the Producer anyways. Our situation is
#   unusual because we have multiple Subchannels that get merged into the
#   same underlying Connection: each Subchannel's Protocol can register a
#   producer on the Subchannel (which is an ITransport), but that adds it to
#   a set of Producers for the Connection (which is also an ITransport). So
#   if the Subchannel is closed, we need to remove its Producer (if any) even
#   though the Connection remains open.
#
# * Register ourselves as an IPushProducer with each successive Connection
#   object. These connections will come and go, but there will never be more
#   than one. When the connection goes away, pause all our producers. When a
#   new one is established, write all our queued messages, then unpause our
#   producers as we would in resumeProducing.
#
# * Inside our resumeProducing call, we'll cycle through all producers,
#   calling their individual resumeProducing methods one at a time. If they
#   write so much data that the Connection pauses us again, we'll find out
#   because our pauseProducing will be called inside that loop. When that
#   happens, we need to stop looping. If we make it through the whole loop
#   without being paused, then all subchannel Producers are left unpaused,
#   and are free to write whenever they want. During this loop, some
#   Producers will be paused, and others will be resumed
#
# * If our pauseProducing is called, all Producers must be paused, and a flag
#   should be set to notify the resumeProducing loop to exit
#
# * In between calls to our resumeProducing method, we're in one of two
#   states.
#   * If we're writing data too fast, then we'll be left in the "paused"
#     state, in which all Subchannel producers are paused, and the aggregate
#     is paused too (our Connection told us to pauseProducing and hasn't yet
#     told us to resumeProducing). In this state, activity is driven by the
#     outbound TCP window opening up, which calls resumeProducing and allows
#     (probably just) one message to be sent. We receive pauseProducing in
#     the middle of their transport.write, so the loop exits early, and the
#     only state change is that some other Producer should get to go next
#     time.
#   * If we're writing too slowly, we'll be left in the "unpaused" state: all
#     Subchannel producers are unpaused, and the aggregate is unpaused too
#     (resumeProducing is the last thing we've been told). In this state,
#     activity is driven by the Subchannels doing a transport.write, which
#     queues some data on the TCP connection (and then might call
#     pauseProducing if it's now full).
#
# * We want to guard against:
#
#   * application protocol registering a Producer without first unregistering
#     the previous one
#
#   * application protocols writing data despite being told to pause
#     (Subchannels without a registered Producer cannot be throttled, and we
#     can't do anything about that, but we must also handle the case where
#     they give us a pause switch and then proceed to ignore it)
#
#   * our Connection calling resumeProducing or pauseProducing without an
#     intervening call of the other kind
#
#   * application protocols that don't handle a resumeProducing or
#     pauseProducing call without an intervening call of the other kind (i.e.
#     we should keep track of the last thing we told them, and not repeat
#     ourselves)
#
# * If the Wormhole is closed, all Subchannels should close. This is not our
#   responsibility: it lives in (Manager? Inbound?)
#
# * If we're given an IPullProducer, we should keep calling its
#   resumeProducing until it runs out of data. We still want fairness, so we
#   won't call it a second time until everyone else has had a turn.


# There are a couple of different ways to approach this. The one I've
# selected is:
#
# * keep a dict that maps from Subchannel to Producer, which only contains
#   entries for Subchannels that have registered a producer. We use this to
#   remove Producers when Subchannels are closed
#
# * keep a Deque of Producers. This represents the fair-throttling rotation:
#   the left-most item gets the next upcoming turn, and then they'll be moved
#   to the end of the queue.
#
# * keep a set of IPushProducers which are paused, a second set of
#   IPushProducers which are unpaused, and a third set of IPullProducers
#   (which are always left paused) Enforce the invariant that these three
#   sets are disjoint, and that their union equals the contents of the deque.
#
# * keep a "paused" flag, which is cleared upon entry to resumeProducing, and
#   set upon entry to pauseProducing. The loop inside resumeProducing checks
#   this flag after each call to producer.resumeProducing, to sense whether
#   they used their turn to write data, and if that write was large enough to
#   fill the TCP window. If set, we break out of the loop. If not, we look
#   for the next producer to unpause. The loop finishes when all producers
#   are unpaused (evidenced by the two sets of paused producers being empty)
#
# * the "paused" flag also determines whether new IPushProducers are added to
#   the paused or unpaused set (IPullProducers are always added to the
#   pull+paused set). If we have any IPullProducers, we're always in the
#   "writing data too fast" state.

# other approaches that I didn't decide to do at this time (but might use in
# the future):
#
# * use one set instead of two. pros: fewer moving parts. cons: harder to
#   spot decoherence bugs like adding a producer to the deque but forgetting
#   to add it to one of the
#
# * use zero sets, and keep the paused-vs-unpaused state in the Subchannel as
#   a visible boolean flag. This conflates Subchannels with their associated
#   Producer (so if we went this way, we should also let them track their own
#   Producer). Our resumeProducing loop ends when 'not any(sc.paused for sc
#   in self._subchannels_with_producers)'. Pros: fewer subchannel->producer
#   mappings lying around to disagree with one another. Cons: exposes a bit
#   too much of the Subchannel internals
#

@attrs
@implementer(IOutbound, IPushProducer)
class Outbound:
    # Manage outbound data: subchannel writes to us, we write to transport
    _manager = attrib(validator=provides(IDilationManager))
    _cooperator = attrib()

    def __attrs_post_init__(self):
        # _outbound_queue holds all messages we've ever sent but not retired
        self._outbound_queue = deque()
        self._next_outbound_seqnum = 0
        # _queued_unsent are messages to retry with our new connection
        self._queued_unsent = deque()

        # outbound flow control: the Connection throttles our writes
        self._subchannel_producers = {}  # Subchannel -> IProducer
        self._paused = True  # our Connection called our pauseProducing
        self._all_producers = deque()  # rotates, left-is-next
        self._paused_producers = set()
        self._unpaused_producers = set()
        self._check_invariants()

        self._connection = None

    def _check_invariants(self):
        assert self._unpaused_producers.isdisjoint(self._paused_producers)
        assert (self._paused_producers.union(self._unpaused_producers) ==
                set(self._all_producers))

    def build_record(self, record_type, *args):
        seqnum = self._next_outbound_seqnum
        self._next_outbound_seqnum += 1
        r = record_type(seqnum, *args)
        assert hasattr(r, "seqnum"), r  # only Open/Data/Close
        return r

    def queue_and_send_record(self, r):
        # we always queue it, to resend on a subsequent connection if
        # necessary
        self._outbound_queue.append(r)

        if self._connection:
            if self._queued_unsent:
                # to maintain correct ordering, queue this instead of sending it
                self._queued_unsent.append(r)
            else:
                # we're allowed to send it immediately
                self._connection.send_record(r)

    def send_if_connected(self, r):
        assert isinstance(r, (KCM, Ping, Pong, Ack)), r  # nothing with seqnum
        if self._connection:
            self._connection.send_record(r)

    # our subchannels call these to register a producer

    def subchannel_registerProducer(self, sc, producer, streaming):
        # streaming==True: IPushProducer (pause/resume)
        # streaming==False: IPullProducer (just resume)
        if sc in self._subchannel_producers:
            raise ValueError(
                "registering producer %s before previous one (%s) was "
                "unregistered" % (producer,
                                  self._subchannel_producers[sc]))
        # our underlying Connection uses streaming==True, so to make things
        # easier, use an adapter when the Subchannel asks for streaming=False
        if not streaming:
            def unregister():
                self.subchannel_unregisterProducer(sc)
            producer = PullToPush(producer, unregister, self._cooperator)

        self._subchannel_producers[sc] = producer
        self._all_producers.append(producer)
        if self._paused:
            self._paused_producers.add(producer)
        else:
            self._unpaused_producers.add(producer)
        self._check_invariants()
        if streaming:
            if self._paused:
                # IPushProducers need to be paused immediately, before they
                # speak
                producer.pauseProducing()  # you wake up sleeping
        else:
            # our PullToPush adapter must be started, but if we're paused then
            # we tell it to pause before it gets a chance to write anything
            producer.startStreaming(self._paused)

    def subchannel_unregisterProducer(self, sc):
        # TODO: what if the subchannel closes, so we unregister their
        # producer for them, then the application reacts to connectionLost
        # with a duplicate unregisterProducer?
        p = self._subchannel_producers.pop(sc)
        if isinstance(p, PullToPush):
            p.stopStreaming()
        self._all_producers.remove(p)
        self._paused_producers.discard(p)
        self._unpaused_producers.discard(p)
        self._check_invariants()

    def subchannel_closed(self, scid, sc):
        self._check_invariants()
        if sc in self._subchannel_producers:
            self.subchannel_unregisterProducer(sc)

    # our Manager tells us when we've got a new Connection to work with

    def use_connection(self, c):
        self._connection = c
        assert not self._queued_unsent
        self._queued_unsent.extend(self._outbound_queue)
        # the connection can tell us to pause when we send too much data
        c.transport.registerProducer(self, True)  # IPushProducer: pause+resume
        # send our queued messages
        self.resumeProducing()

    def stop_using_connection(self):
        self._connection.transport.unregisterProducer()
        self._connection = None
        self._queued_unsent.clear()
        self.pauseProducing()
        # TODO: I expect this will call pauseProducing twice: the first time
        # when we get stopProducing (since we're registered with the
        # underlying connection as the producer), and again when the manager
        # notices the connectionLost and calls our _stop_using_connection

    def handle_ack(self, resp_seqnum):
        # we've received an inbound ack, so retire something
        while (self._outbound_queue and
               self._outbound_queue[0].seqnum <= resp_seqnum):
            self._outbound_queue.popleft()
        while (self._queued_unsent and
               self._queued_unsent[0].seqnum <= resp_seqnum):
            self._queued_unsent.popleft()
        # Inbound is responsible for tracking the high watermark and deciding
        # whether to ignore inbound messages or not

    # IPushProducer: the active connection calls these because we used
    # c.transport.registerProducer to ask for them

    def pauseProducing(self):
        if self._paused:
            return  # someone is confused and called us twice
        self._paused = True
        for p in self._all_producers:
            if p in self._unpaused_producers:
                self._unpaused_producers.remove(p)
                self._paused_producers.add(p)
                p.pauseProducing()

    def resumeProducing(self):
        if not self._paused:
            return  # someone is confused and called us twice
        self._paused = False

        while not self._paused:
            if self._queued_unsent:
                r = self._queued_unsent.popleft()
                self._connection.send_record(r)
                continue
            p = self._get_next_unpaused_producer()
            if not p:
                break
            self._paused_producers.remove(p)
            self._unpaused_producers.add(p)
            p.resumeProducing()

    def _get_next_unpaused_producer(self):
        self._check_invariants()
        if not self._paused_producers:
            return None
        while True:
            p = self._all_producers[0]
            self._all_producers.rotate(-1)  # p moves to the end of the line
            # the only unpaused Producers are at the end of the list
            assert p in self._paused_producers
            return p

    def stopProducing(self):
        # we'll hopefully have a new connection to work with in the future,
        # so we don't shut anything down. We do pause everyone, though.
        self.pauseProducing()


# modelled after twisted.internet._producer_helper._PullToPush , but with a
# configurable Cooperator, a pause-immediately argument to startStreaming()
@implementer(IPushProducer)
@attrs(eq=False)
class PullToPush:
    _producer = attrib(validator=provides(IPullProducer))
    _unregister = attrib(validator=lambda _a, _b, v: callable(v))
    _cooperator = attrib()
    _finished = False

    def _pull(self):
        while True:
            try:
                self._producer.resumeProducing()
            except Exception:
                log.err(None, "%s failed, producing will be stopped:" %
                        (safe_str(self._producer),))
                try:
                    self._unregister()
                    # The consumer should now call stopStreaming() on us,
                    # thus stopping the streaming.
                except Exception:
                    # Since the consumer blew up, we may not have had
                    # stopStreaming() called, so we just stop on our own:
                    log.err(None, "%s failed to unregister producer:" %
                            (safe_str(self._unregister),))
                    self._finished = True
                    return
            yield None

    def startStreaming(self, paused):
        self._coopTask = self._cooperator.cooperate(self._pull())
        if paused:
            self.pauseProducing()  # timer is scheduled, but task is removed

    def stopStreaming(self):
        if self._finished:
            return
        self._finished = True
        self._coopTask.stop()

    def pauseProducing(self):
        self._coopTask.pause()

    def resumeProducing(self):
        self._coopTask.resume()

    def stopProducing(self):
        self.stopStreaming()
        self._producer.stopProducing()
